# coding: utf-8
import io
import json
import os
import re
import struct
from collections import defaultdict, namedtuple
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy
from tqdm import tqdm

import supervisely as sly
import supervisely.volume_annotation.constants as volume_constants
from supervisely._utils import batched
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.project_api import ProjectInfo
from supervisely.api.volume.volume_api import VolumeInfo
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import change_directory_at_index, touch
from supervisely.project.project import OpenMode
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_type import ProjectType
from supervisely.project.video_project import VideoDataset, VideoProject
from supervisely.project.versioning.common import (
    DEFAULT_VOLUME_SCHEMA_VERSION,
    get_volume_snapshot_schema,
)
from supervisely.project.versioning.schema_fields import VersionSchemaField
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress, tqdm_sly
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume import stl_converter
from supervisely.volume import volume as sly_volume
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_figure import VolumeFigure

VolumeItemPaths = namedtuple("VolumeItemPaths", ["volume_path", "ann_path"])


class VolumeDataset(VideoDataset):
    item_dir_name = "volume"
    interpolation_dir = "interpolation"
    interpolation_dir_name = interpolation_dir
    mask_dir = "mask"
    mask_dir_name = mask_dir
    annotation_class = VolumeAnnotation
    item_module = sly_volume
    paths_tuple = VolumeItemPaths

    @classmethod
    def _has_valid_ext(cls, path: str) -> bool:
        """
        Checks if file from given path is supported
        :param path: str
        :return: bool
        """
        return sly_volume.has_valid_ext(path)

    def _get_empty_annotaion(self, item_name):
        path = item_name
        _, volume_meta = sly_volume.read_nrrd_serie_volume(path)
        return self.annotation_class(volume_meta)

    def get_interpolation_dir(self, item_name):
        return os.path.join(self.directory, self.interpolation_dir, item_name)

    def get_interpolation_path(self, item_name, figure):
        return os.path.join(self.get_interpolation_dir(item_name), figure.key().hex + ".stl")

    def get_mask_dir(self, item_name):
        return os.path.join(self.directory, self.mask_dir, item_name)

    def get_mask_path(self, item_name, figure):
        return os.path.join(self.get_mask_dir(item_name), figure.key().hex + ".nrrd")

    def get_classes_stats(
        self,
        project_meta: Optional[ProjectMeta] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        if project_meta is None:
            project = VolumeProject(self.project_dir, OpenMode.READ)
            project_meta = project.meta
        class_items = {}
        class_objects = {}
        class_figures = {}
        for obj_class in project_meta.obj_classes:
            class_items[obj_class.name] = 0
            class_objects[obj_class.name] = 0
            class_figures[obj_class.name] = 0
        for item_name in self:
            item_ann = self.get_ann(item_name, project_meta)
            item_class = {}
            for ann_obj in item_ann.objects:
                class_objects[ann_obj.obj_class.name] += 1
            for volume_figure in item_ann.figures:
                class_figures[volume_figure.parent_object.obj_class.name] += 1
                item_class[volume_figure.parent_object.obj_class.name] = True
            for obj_class in project_meta.obj_classes:
                if obj_class.name in item_class.keys():
                    class_items[obj_class.name] += 1

        result = {}
        if return_items_count:
            result["items_count"] = class_items
        if return_objects_count:
            result["objects_count"] = class_objects
        if return_figures_count:
            result["figures_count"] = class_figures
        return result


class VolumeProject(VideoProject):
    dataset_class = VolumeDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = VolumeDataset

    _SERIALIZATION_MAGIC = b"SLYVOLPAR"
    _SERIALIZATION_VERSION = 1
    _SECTION_PROJECT_INFO = 1
    _SECTION_PROJECT_META = 2
    _SECTION_DATASETS = 3
    _SECTION_VOLUMES = 4
    _SECTION_ANNOTATIONS = 5

    def get_classes_stats(
        self,
        dataset_names: Optional[List[str]] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        return super(VolumeProject, self).get_classes_stats(
            dataset_names, return_objects_count, return_figures_count, return_items_count
        )

    @property
    def type(self) -> str:
        """
        Project type.

        :return: Project type.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            project = sly.VolumeProject("/home/admin/work/supervisely/projects/volumes", sly.OpenMode.READ)
            print(project.type)
            # Output: 'volumes'
        """
        return ProjectType.VOLUMES.value

    @staticmethod
    def download(
        api: Api,
        project_id: int,
        dest_dir: str,
        dataset_ids: Optional[List[int]] = None,
        download_volumes: Optional[bool] = True,
        log_progress: bool = False,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        **kwargs,
    ) -> None:
        """
        Download volume project from Supervisely to the given directory.

        :param api: Supervisely API address and token.
        :type api: :class:`Api<supervisely.api.api.Api>`
        :param project_id: Supervisely downloadable project ID.
        :type project_id: :class:`int`
        :param dest_dir: Destination directory.
        :type dest_dir: :class:`str`
        :param dataset_ids: Dataset IDs.
        :type dataset_ids: :class:`list` [ :class:`int` ], optional
        :param download_volumes: Download volume data files or not.
        :type download_volumes: :class:`bool`, optional
        :param log_progress: Show uploading progress bar.
        :type log_progress: bool
        :param progress_cb: Function for tracking the download progress.
        :type progress_cb: tqdm or callable, optional

        :return: None
        :rtype: NoneType
        :Usage example:
        .. code-block:: python

                import supervisely as sly

                # Local destination Volume Project folder
                save_directory = "/home/admin/work/supervisely/source/vlm_project"

                # Obtain server address and your api_token from environment variables
                # Edit those values if you run this notebook on your own PC
                address = os.environ['SERVER_ADDRESS']
                token = os.environ['API_TOKEN']

                # Initialize API object
                api = sly.Api(address, token)
                project_id = 8888

                # Download Project
                sly.VolumeProject.download(api, project_id, save_directory)
                project_fs = sly.VolumeProject(save_directory, sly.OpenMode.READ)
        """
        download_volume_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            download_volumes=download_volumes,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )

    @staticmethod
    def download_bin(
        api: Api,
        project_id: int,
        dest_dir: Optional[str] = None,
        dataset_ids: Optional[List[int]] = None,
        download_volumes: bool = True,
        log_progress: bool = False,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        return_bytesio: bool = False,
        schema_version: str = DEFAULT_VOLUME_SCHEMA_VERSION,
        *args,
        **kwargs,
    ) -> Union[str, io.BytesIO]:
        """
        Download a Volume Project snapshot into a Parquet-backed binary blob (`.tar.zst` file or in-memory BytesIO).

        The snapshot stores:

        - Project info and meta
        - Dataset tree (dataset infos)
        - Volume infos (optionally)
        - Volume annotations (for the included volumes)

        The resulting binary snapshot can be restored later with :func:`upload_bin`.

        :param api: Supervisely API client.
        :type api: :class:`~supervisely.api.api.Api`
        :param project_id: Source Volume Project ID on the server.
        :type project_id: int
        :param dest_dir: Local folder where the snapshot file will be written. Required when `return_bytesio=False`.
        :type dest_dir: str, optional
        :param dataset_ids: Optional list of dataset IDs to include. If provided, only these datasets will be included (recursively, preserving tree structure where applicable).
        :type dataset_ids: List[int], optional
        :param download_volumes: If False, only project/meta/dataset tree is stored (volume infos and annotations are skipped). This is useful for “structure-only” snapshots.
        :type download_volumes: bool, optional
        :param log_progress: If True, show a progress bar (unless a custom ``progress_cb`` is provided).
        :type log_progress: bool
        :param progress_cb: Optional callback (or tqdm-like object) called with incremental progress.
        :type progress_cb: tqdm or callable, optional
        :param return_bytesio: If True, return an in-memory :class:`io.BytesIO` with snapshot bytes. If False, write snapshot to ``dest_dir`` and return the file path.
        :type return_bytesio: bool, optional
        :param schema_version: Snapshot schema version. Controls the internal Parquet layout/fields. Supported values are the keys from :func:`~supervisely.project.volume_schema.get_volume_snapshot_schema` (currently: ``"v2.0.0"``).
        :type schema_version: str, optional
        :return: Snapshot file path (when ``return_bytesio=False``) or a BytesIO (when ``return_bytesio=True``).
        :rtype: str or io.BytesIO
        :raises ValueError: If ``dest_dir`` is not provided and ``return_bytesio`` is False.
        :raises RuntimeError: If required optional dependencies (e.g. pyarrow) are missing.

        :Usage example:

        .. code-block:: python

            import supervisely as sly
            import os

            api = sly.Api(os.environ["SERVER_ADDRESS"], os.environ["API_TOKEN"])

            # 1) Save snapshot to disk
            out_path = sly.VolumeProject.download_bin(
                api,
                project_id=123,
                dest_dir="/tmp/vol_project_snapshot",
                download_volumes=True,
                log_progress=True,
            )

            # 2) Create an in-memory snapshot (BytesIO) and restore it
            blob = sly.VolumeProject.download_bin(
                api,
                project_id=123,
                return_bytesio=True,
                download_volumes=False,  # structure-only
            )
            restored = sly.VolumeProject.upload_bin(api, blob, workspace_id=45, project_name="Restored")
        """

        pa = VolumeProject._require_pyarrow()
        snapshot_schema = get_volume_snapshot_schema(schema_version)

        if dest_dir is None and not return_bytesio:
            raise ValueError(
                "Local save directory dest_dir must be specified if return_bytesio is False"
            )

        ds_filters = (
            [{"field": "id", "operator": "in", "value": dataset_ids}]
            if dataset_ids is not None
            else None
        )

        project_info = api.project.get_info_by_id(project_id)
        project_meta = api.project.get_meta(project_id, with_settings=True)
        project_meta_obj = ProjectMeta.from_json(project_meta)
        dataset_infos = api.dataset.get_list(project_id, filters=ds_filters, recursive=True, include_custom_data=True)

        dataset_records = [dataset_info._asdict() for dataset_info in dataset_infos]
        volume_records: List[Dict] = []
        annotations: Dict[str, Dict] = {}
        key_id_map = KeyIdMap()

        for dataset_info in dataset_infos:
            if dataset_ids is not None and dataset_info.id not in dataset_ids:
                continue

            volumes = api.volume.get_list(dataset_info.id)
            if len(volumes) == 0:
                continue

            if not download_volumes:
                continue

            ds_progress = progress_cb
            if log_progress and progress_cb is None:
                ds_progress = tqdm_sly(
                    desc="Collecting volumes from: {!r}".format(dataset_info.name),
                    total=len(volumes),
                )

            volume_ids = [volume_info.id for volume_info in volumes]
            ann_jsons = api.volume.annotation.download_bulk(dataset_info.id, volume_ids)

            # insert custom_data into ann_jsons (api does not return it in download_bulk atm)
            # Build mappings:
            # - volume_id -> ann_json
            # - volume_id -> {figure_id -> spatial_figure_dict}
            ann_by_volume_id: Dict[int, Dict[str, Any]] = {}
            spatial_figures_by_volume: Dict[int, Dict[int, Dict[str, Any]]] = {}
            for ann_json in ann_jsons:
                volume_id = ann_json.get(ApiField.VOLUME_ID)
                if volume_id is None:
                    continue
                ann_by_volume_id[volume_id] = ann_json
                figures_list = ann_json.get(volume_constants.SPATIAL_FIGURES, []) or []
                fig_id_to_spatial_figure: Dict[int, Dict[str, Any]] = {}
                for spatial_figure in figures_list:
                    fig_id = spatial_figure.get("id")
                    if fig_id is not None:
                        fig_id_to_spatial_figure[fig_id] = spatial_figure
                spatial_figures_by_volume[volume_id] = fig_id_to_spatial_figure

            figures_dict = api.volume.figure.download(dataset_info.id, volume_ids)
            for volume_id, figure_infos in figures_dict.items():
                ann_json = ann_by_volume_id.get(volume_id)
                if ann_json is None:
                    continue
                fig_id_to_spatial_figure = spatial_figures_by_volume.get(volume_id, {})
                for figure_info in figure_infos:
                    spatial_figure = fig_id_to_spatial_figure.get(figure_info.id)
                    if spatial_figure is not None:
                        spatial_figure[ApiField.CUSTOM_DATA] = figure_info.custom_data

            for volume_info, ann_json in zip(volumes, ann_jsons):
                ann_dict = snapshot_schema.annotation_dict_from_raw(
                    api=api,
                    raw_ann_json=ann_json,
                    project_meta_obj=project_meta_obj,
                    key_id_map=key_id_map,
                )
                volume_records.append(volume_info._asdict())
                annotations[str(volume_info.id)] = ann_dict
                if progress_cb is not None:
                    progress_cb(1)
                if ds_progress is not None:
                    ds_progress(1)

        project_info_dict = project_info._asdict()
        project_info_dict[VersionSchemaField.SCHEMA_VERSION] = schema_version
        payload = {
            "project_info": project_info_dict,
            "project_meta": project_meta,
            "dataset_infos": dataset_records,
            "volume_infos": volume_records,
            "annotations": annotations,
        }
        blob = VolumeProject._serialize_payload_to_parquet_blob(pa, payload, snapshot_schema)

        if return_bytesio:
            stream = io.BytesIO(blob)
            stream.seek(0)
            return stream

        os.makedirs(dest_dir, exist_ok=True)
        file_path = os.path.join(dest_dir, f"{project_info.id}_{project_info.name}.arrow")
        with open(file_path, "wb") as out:
            out.write(blob)
        return file_path

    @staticmethod
    def upload(
        directory: str,
        api: Api,
        workspace_id: int,
        project_name: Optional[str] = None,
        log_progress: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Tuple[int, str]:
        """
        Uploads volume project to Supervisely from the given directory.

        :param directory: Path to project directory.
        :type directory: :class:`str`
        :param api: Supervisely API address and token.
        :type api: :class:`Api<supervisely.api.api.Api>`
        :param workspace_id: Workspace ID, where project will be uploaded.
        :type workspace_id: :class:`int`
        :param project_name: Name of the project in Supervisely. Can be changed if project with the same name is already exists.
        :type project_name: :class:`str`, optional
        :param log_progress: Show uploading progress bar.
        :type log_progress: :class:`bool`
        :param progress_cb: Function for tracking the download progress.
        :type progress_cb: tqdm or callable, optional

        :return: Project ID and name. It is recommended to check that returned project name coincides with provided project name.
        :rtype: :class:`int`, :class:`str`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            # Local folder with Volume Project
            project_directory = "/home/admin/work/supervisely/source/vlm_project"

            # Obtain server address and your api_token from environment variables
            # Edit those values if you run this notebook on your own PC
            address = os.environ['SERVER_ADDRESS']
            token = os.environ['API_TOKEN']

            # Initialize API object
            api = sly.Api(address, token)

            # Upload Volume Project
            project_id, project_name = sly.VolumeProject.upload(
                project_directory,
                api,
                workspace_id=45,
                project_name="My Volume Project"
            )
        """
        return upload_volume_project(
            dir=directory,
            api=api,
            workspace_id=workspace_id,
            project_name=project_name,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )

    @staticmethod
    def upload_bin(
        api: Api,
        file: Union[str, io.BytesIO],
        workspace_id: int,
        project_name: Optional[str] = None,
        log_progress: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_missed_entities: bool = False,
        *args,
        **kwargs,
    ) -> ProjectInfo:
        """
        Restore a volume project from a Parquet blob produced by :func:`download_bin`.

        :param api: Supervisely API client.
        :type api: :class:`~supervisely.api.api.Api`
        :param file: Snapshot file path (``.tar.zst``) or an in-memory :class:`io.BytesIO` stream.
        :type file: Union[str, io.BytesIO]
        :param workspace_id: Target workspace ID where the project will be created.
        :type workspace_id: int
        :param project_name: Optional new project name. If not provided, the name from the snapshot will be used. If the name already exists in the workspace, a free name will be chosen.
        :type project_name: str, optional
        :param log_progress: If True, show a progress bar (unless a custom ``progress_cb`` is provided).
        :type log_progress: bool
        :param progress_cb: Optional callback (or tqdm-like object) called with incremental progress.
        :type progress_cb: tqdm or callable, optional
        :param skip_missed_entities: If True, skip volumes that cannot be restored because their source hash is missing in the snapshot payload. If False, such cases raise an error.
        :type skip_missed_entities: bool
        :return: Info of the newly created project.
        :rtype: :class:`~supervisely.api.project_api.ProjectInfo`
        :raises RuntimeError: If the snapshot contains volumes without hashes and ``skip_missed_entities`` is False.
        """

        pa = VolumeProject._require_pyarrow()

        if isinstance(file, io.BytesIO):
            raw_data = file.getbuffer()
        else:
            with open(file, "rb") as src:
                raw_data = src.read()

        payload = VolumeProject._deserialize_payload_from_parquet(pa, raw_data)

        project_meta = ProjectMeta.from_json(payload["project_meta"])
        project_info: Dict = payload.get("project_info", {})
        dataset_records: List[Dict] = payload.get("dataset_infos", [])
        volume_records: List[Dict] = payload.get("volume_infos", [])
        annotations: Dict[str, Dict] = payload.get("annotations", {})

        project_title = project_name or project_info.get("name")
        if api.project.exists(workspace_id, project_title):
            project_title = api.project.get_free_name(workspace_id, project_title)
        src_project_desc = project_info.get("description")
        new_project_info = api.project.create(
            workspace_id,
            project_title,
            ProjectType.VOLUMES,
            description=src_project_desc,
            readme=project_info.get("readme"),
        )
        api.project.update_meta(new_project_info.id, project_meta)

        custom_data = new_project_info.custom_data
        source_project_id = payload["project_info"].get("id")
        version_info = payload["project_info"].get("version") or {}
        custom_data["restored_from"] = {
            "project_id": source_project_id,
            "version_num": version_info.get("version"),
        }
        original_custom_data = payload["project_info"].get("custom_data") or {}
        custom_data.update(original_custom_data)
        api.project.update_custom_data(new_project_info.id, custom_data, silent=True)

        dataset_mapping: Dict[int, sly.DatasetInfo] = {}
        sorted_datasets = sorted(
            dataset_records,
            key=lambda data: (data.get("parent_id") is not None, data.get("parent_id") or 0),
        )
        for dataset_data in sorted_datasets:
            parent_ds_info = dataset_mapping.get(dataset_data.get("parent_id"))
            new_parent_id = parent_ds_info.id if parent_ds_info else None
            new_dataset_info = api.dataset.create(
                project_id=new_project_info.id,
                name=dataset_data.get("name"),
                description=dataset_data.get("description"),
                parent_id=new_parent_id,
                custom_data=dataset_data.get("custom_data"),
            )
            dataset_mapping[dataset_data.get("id")] = new_dataset_info

        volume_mapping: Dict[int, VolumeInfo] = {}
        volumes_by_dataset: Dict[int, List[Dict]] = defaultdict(list)
        for volume_data in volume_records:
            volumes_by_dataset[volume_data.get("dataset_id")].append(volume_data)

        for old_dataset_id, dataset_volumes in volumes_by_dataset.items():
            new_dataset_info = dataset_mapping.get(old_dataset_id)
            if new_dataset_info is None:
                continue

            dataset_volumes_to_upload: List[Dict] = []
            missing_names: List[str] = []
            for vol in dataset_volumes:
                if vol.get("hash"):
                    dataset_volumes_to_upload.append(vol)
                else:
                    missing_names.append(vol.get("name") or str(vol.get("id")))

            if missing_names:
                if skip_missed_entities:
                    for vol_name in missing_names:
                        logger.warning(
                            "Volume %r skipped during restoration because its source hash is unavailable.",
                            vol_name,
                        )
                    if len(dataset_volumes_to_upload) == 0:
                        continue
                else:
                    raise RuntimeError(
                        "Cannot restore volumes without available hash. Missing volume names: {}".format(
                            ", ".join(missing_names)
                        )
                    )

            hashes = [volume.get("hash") for volume in dataset_volumes_to_upload]
            names = [volume.get("name") for volume in dataset_volumes_to_upload]
            metas = [volume.get("meta") for volume in dataset_volumes_to_upload]

            ds_progress = progress_cb
            if log_progress and progress_cb is None:
                ds_progress = tqdm_sly(
                    desc="Uploading volumes to {!r}".format(new_dataset_info.name),
                    total=len(dataset_volumes_to_upload),
                )

            new_volume_infos = api.volume.upload_hashes(
                new_dataset_info.id,
                names=names,
                hashes=hashes,
                metas=metas,
                progress_cb=ds_progress,
            )

            for old_volume, new_volume in zip(dataset_volumes_to_upload, new_volume_infos):
                volume_mapping[old_volume.get("id")] = new_volume

        for volume_id_str, ann_json in annotations.items():
            new_volume_info = volume_mapping.get(int(volume_id_str))
            if new_volume_info is None:
                if skip_missed_entities:
                    logger.warning(
                        "Annotation for volume %s skipped because the source volume was not restored.",
                        volume_id_str,
                    )
                continue
            ann_json["volumeId"] = new_volume_info.id
            ann = VolumeAnnotation.from_json(ann_json, project_meta, None)
            api.volume.annotation.append(new_volume_info.id, ann, None)

        return api.project.get_info_by_id(new_project_info.id)

    @staticmethod
    def _require_pyarrow():
        try:
            import pyarrow as pa  # pylint: disable=import-error
        except ModuleNotFoundError as exc:
            raise ModuleNotFoundError(
                "VolumeProject binary versioning requires the optional dependency 'pyarrow'. "
                "Install it with `pip install pyarrow` to use download_bin/upload_bin."
            ) from exc
        return pa

    @staticmethod
    def _serialize_payload_to_parquet_blob(pa_module, payload: Dict[str, Dict], snapshot_schema) -> bytes:
        dataset_records: List[Dict] = payload.get("dataset_infos", []) or []
        volume_records: List[Dict] = payload.get("volume_infos", []) or []
        annotations_dict: Dict[str, Dict] = payload.get("annotations", {}) or {}

        dataset_rows = [snapshot_schema.dataset_row_from_record(r) for r in dataset_records]
        dataset_table = pa_module.Table.from_pylist(
            dataset_rows, schema=snapshot_schema.datasets_table_schema(pa_module)
        )

        volume_rows = [snapshot_schema.volume_row_from_record(r) for r in volume_records]
        volume_table = pa_module.Table.from_pylist(
            volume_rows, schema=snapshot_schema.volumes_table_schema(pa_module)
        )

        ann_rows = []
        for volume_id_str, ann in annotations_dict.items():
            try:
                src_volume_id = int(volume_id_str)
            except (TypeError, ValueError):
                continue
            ann_rows.append(
                snapshot_schema.annotation_row_from_dict(src_volume_id=src_volume_id, annotation=ann)
            )
        annotations_table = pa_module.Table.from_pylist(
            ann_rows, schema=snapshot_schema.annotations_table_schema(pa_module)
        )

        sections = [
            (
                VolumeProject._SECTION_PROJECT_INFO,
                VolumeProject._json_bytes(payload.get("project_info", {})),
            ),
            (
                VolumeProject._SECTION_PROJECT_META,
                VolumeProject._json_bytes(payload.get("project_meta", {})),
            ),
            (
                VolumeProject._SECTION_DATASETS,
                VolumeProject._table_to_parquet_bytes(pa_module, dataset_table),
            ),
            (
                VolumeProject._SECTION_VOLUMES,
                VolumeProject._table_to_parquet_bytes(pa_module, volume_table),
            ),
            (
                VolumeProject._SECTION_ANNOTATIONS,
                VolumeProject._table_to_parquet_bytes(pa_module, annotations_table),
            ),
        ]

        return VolumeProject._assemble_sections(sections)

    @staticmethod
    def _build_table(pa_module, columns: Dict[str, Tuple[List, Any]]):
        arrays = {}
        for name, (values, dtype) in columns.items():
            arrays[name] = pa_module.array(values, type=dtype)
        return pa_module.table(arrays)

    @staticmethod
    def _table_to_parquet_bytes(pa_module, table) -> bytes:
        from pyarrow import parquet as pq  # pylint: disable=import-error

        sink = pa_module.BufferOutputStream()
        pq.write_table(table, sink)
        return sink.getvalue().to_pybytes()

    @staticmethod
    def _parquet_bytes_to_table(pa_module, data: bytes):
        if not data:
            return pa_module.table({})
        from pyarrow import parquet as pq  # pylint: disable=import-error

        buffer = pa_module.BufferReader(data)
        return pq.read_table(buffer)

    @staticmethod
    def _json_dumps(data) -> str:
        if isinstance(data, str):
            return data
        return json.dumps(data, ensure_ascii=False)

    @staticmethod
    def _json_bytes(data) -> bytes:
        return VolumeProject._json_dumps(data).encode("utf-8")

    @staticmethod
    def _assemble_sections(sections: List[Tuple[int, bytes]]) -> bytes:
        if len(sections) > 255:
            raise RuntimeError("Too many sections for VolumeProject binary payload")
        buffer = io.BytesIO()
        buffer.write(VolumeProject._SERIALIZATION_MAGIC)
        buffer.write(struct.pack(">B", VolumeProject._SERIALIZATION_VERSION))
        buffer.write(struct.pack(">B", len(sections)))
        for section_type, payload in sections:
            if payload is None:
                payload = b""
            buffer.write(struct.pack(">B", section_type))
            buffer.write(struct.pack(">Q", len(payload)))
            buffer.write(payload)
        return buffer.getvalue()

    @staticmethod
    def _parse_parquet_sections(raw_data) -> Dict[int, bytes]:
        magic = VolumeProject._SERIALIZATION_MAGIC
        view = raw_data if isinstance(raw_data, memoryview) else memoryview(raw_data)
        header_len = len(magic) + 2
        if len(view) < header_len:
            logger.warning(
                f"VolumeProject binary payload too small: {len(view)} bytes (need >= {header_len}). First bytes(hex)={view[: min(len(view), 16)].tobytes().hex()}",
            )
            raise RuntimeError("Corrupted VolumeProject binary payload")
        if view[: len(magic)].tobytes() != magic:
            found = view[: len(magic)].tobytes()
            logger.warning(
                f"VolumeProject binary payload magic mismatch. expected={magic.hex()} found={found.hex()} total_bytes={len(view)} prefix16(hex)={view[:16].tobytes().hex()}",
            )
            raise RuntimeError(
                "Unsupported VolumeProject binary payload format (magic mismatch). "
                "Expected magic={!r}, found={!r}".format(magic, found)
            )

        offset = len(magic)
        version = view[offset]
        offset += 1
        if version != VolumeProject._SERIALIZATION_VERSION:
            logger.warning(
                "VolumeProject binary payload version mismatch. expected=%d found=%d total_bytes=%d",
                VolumeProject._SERIALIZATION_VERSION,
                version,
                len(view),
            )
            raise RuntimeError(
                "Unsupported VolumeProject binary payload version: {}".format(version)
            )

        section_count = view[offset]
        offset += 1
        sections: Dict[int, bytes] = {}
        for _ in range(section_count):
            if offset + 9 > len(view):
                raise RuntimeError("Corrupted VolumeProject binary payload")
            section_type = view[offset]
            offset += 1
            length = int.from_bytes(view[offset : offset + 8], "big")
            offset += 8
            if offset + length > len(view):
                raise RuntimeError("Corrupted VolumeProject binary payload")
            sections[section_type] = view[offset : offset + length].tobytes()
            offset += length
        return sections

    @staticmethod
    def _deserialize_payload_from_parquet(pa_module, raw_data) -> Dict:
        sections = VolumeProject._parse_parquet_sections(raw_data)

        try:
            project_info = json.loads(sections[VolumeProject._SECTION_PROJECT_INFO].decode("utf-8"))
            project_meta = json.loads(sections[VolumeProject._SECTION_PROJECT_META].decode("utf-8"))
        except KeyError as exc:
            raise RuntimeError("VolumeProject payload missing metadata section") from exc

        if VolumeProject._SECTION_DATASETS not in sections:
            logger.warning("VolumeProject blob has no datasets section; treating as empty.")
        if VolumeProject._SECTION_VOLUMES not in sections:
            logger.warning("VolumeProject blob has no volumes section; treating as empty.")
        if VolumeProject._SECTION_ANNOTATIONS not in sections:
            logger.warning("VolumeProject blob has no annotations section; treating as empty.")

        dataset_table = VolumeProject._parquet_bytes_to_table(
            pa_module, sections.get(VolumeProject._SECTION_DATASETS, b"")
        )
        volume_table = VolumeProject._parquet_bytes_to_table(
            pa_module, sections.get(VolumeProject._SECTION_VOLUMES, b"")
        )
        annotations_table = VolumeProject._parquet_bytes_to_table(
            pa_module, sections.get(VolumeProject._SECTION_ANNOTATIONS, b"")
        )

        dataset_records: List[Dict] = []
        if dataset_table is not None and dataset_table.num_rows:
            col_json = (
                VersionSchemaField.JSON
                if VersionSchemaField.JSON in dataset_table.column_names
                else "json"
            )
            if col_json in dataset_table.column_names:
                dataset_jsons = dataset_table.column(col_json).to_pylist()
                dataset_records = [json.loads(item) for item in dataset_jsons]

        volume_records: List[Dict] = []
        if volume_table is not None and volume_table.num_rows:
            col_json = (
                VersionSchemaField.JSON
                if VersionSchemaField.JSON in volume_table.column_names
                else "json"
            )
            if col_json in volume_table.column_names:
                volume_jsons = volume_table.column(col_json).to_pylist()
                volume_records = [json.loads(item) for item in volume_jsons]

        annotations: Dict[str, Dict] = {}
        if annotations_table is not None and annotations_table.num_rows:
            col_vol_id = (
                VersionSchemaField.SRC_VOLUME_ID
                if VersionSchemaField.SRC_VOLUME_ID in annotations_table.column_names
                else "volume_id"
            )
            col_ann = (
                VersionSchemaField.ANNOTATION
                if VersionSchemaField.ANNOTATION in annotations_table.column_names
                else "annotation"
            )
            if col_vol_id in annotations_table.column_names and col_ann in annotations_table.column_names:
                annotation_ids = annotations_table.column(col_vol_id).to_pylist()
                annotation_payloads = annotations_table.column(col_ann).to_pylist()
                for volume_id, annotation_json in zip(annotation_ids, annotation_payloads):
                    if volume_id is None:
                        continue
                    annotations[str(volume_id)] = json.loads(annotation_json)

        return {
            "project_info": project_info,
            "project_meta": project_meta,
            "dataset_infos": dataset_records,
            "volume_infos": volume_records,
            "annotations": annotations,
        }

    @staticmethod
    def _load_mask_geometries(api: Api, ann: VolumeAnnotation, key_id_map: KeyIdMap) -> None:
        for sf in ann.spatial_figures:
            if sf.geometry.name() != Mask3D.name():
                continue
            api.volume.figure.load_sf_geometry(sf, key_id_map)

    @staticmethod
    def get_train_val_splits_by_count(project_dir: str, train_count: int, val_count: int) -> None:
        """
        Not available for VolumeProject class.
        :raises: :class:`NotImplementedError` in all cases.
        """
        raise NotImplementedError(
            "Static method 'get_train_val_splits_by_count()' is not supported for VolumeProject class now."
        )

    @staticmethod
    def get_train_val_splits_by_tag(
        project_dir: str,
        train_tag_name: str,
        val_tag_name: str,
        untagged: Optional[str] = "ignore",
    ) -> None:
        """
        Not available for VolumeProject class.
        :raises: :class:`NotImplementedError` in all cases.
        """
        raise NotImplementedError(
            "Static method 'get_train_val_splits_by_tag()' is not supported for VolumeProject class now."
        )

    @staticmethod
    def get_train_val_splits_by_dataset(
        project_dir: str, train_datasets: List[str], val_datasets: List[str]
    ) -> None:
        """
        Not available for VolumeProject class.
        :raises: :class:`NotImplementedError` in all cases.
        """
        raise NotImplementedError(
            f"Static method 'get_train_val_splits_by_tag()' is not supported for VolumeProject class now."
        )

    @staticmethod
    def get_train_val_splits_by_collections(
        project_dir: str,
        train_collections: List[int],
        val_collections: List[int],
        project_id: int,
        api: Api,
    ) -> None:
        """
        Not available for VolumeProject class.
        :raises: :class:`NotImplementedError` in all cases.
        """
        raise NotImplementedError(
            f"Static method 'get_train_val_splits_by_collections()' is not supported for VolumeProject class now."
        )

    @staticmethod
    async def download_async(*args, **kwargs):
        raise NotImplementedError(
            f"Static method 'download_async()' is not supported for VolumeProject class now."
        )


def download_volume_project(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    download_volumes: Optional[bool] = True,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> None:
    """
    Download volume project to the local directory.

    :param api: Supervisely API address and token.
    :type api: Api
    :param project_id: Project ID to download.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded. Datasets could be downloaded from different projects but with the same data type.
    :type dataset_ids: list(int), optional
    :param download_volumes: Include volumes in the download.
    :type download_volumes: bool, optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool
    :param progress_cb: Function for tracking download progress.
    :type progress_cb: tqdm or callable, optional

    :return: None.
    :rtype: NoneType
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        from tqdm import tqdm
        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        dest_dir = 'your/local/dest/dir'

        # Download volume project
        project_id = 18532
        project_info = api.project.get_info_by_id(project_id)
        num_volumes = project_info.items_count

        p = tqdm(desc="Downloading volume project", total=num_volumes)
        sly.download_volume_project(
            api,
            project_id,
            dest_dir,
            progress_cb=p,
        )
    """

    LOG_BATCH_SIZE = 1

    key_id_map = KeyIdMap()

    project_fs = VolumeProject(dest_dir, OpenMode.CREATE)

    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    if progress_cb is not None:
        log_progress = False

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs: VolumeDataset = project_fs.create_dataset(dataset.name)
        volumes = api.volume.get_list(dataset.id)

        ds_progress = progress_cb
        if log_progress:
            ds_progress = tqdm_sly(
                desc="Downloading volumes from: {!r}".format(dataset.name),
                total=len(volumes),
            )
        for batch in batched(volumes, batch_size=LOG_BATCH_SIZE):
            volume_ids = [volume_info.id for volume_info in batch]
            volume_names = [volume_info.name for volume_info in batch]

            ann_jsons = api.volume.annotation.download_bulk(dataset.id, volume_ids)

            for volume_id, volume_name, volume_info, ann_json in zip(
                volume_ids, volume_names, batch, ann_jsons
            ):
                if volume_name != ann_json[ApiField.VOLUME_NAME]:
                    raise RuntimeError(
                        "Error in api.volume.annotation.download_batch: broken order"
                    )
                try:
                    ann = VolumeAnnotation.from_json(ann_json, project_fs.meta, key_id_map)
                except Exception as e:
                    logger.info(
                        "INFO FOR DEBUGGING",
                        extra={
                            "project_id": project_id,
                            "dataset_id": dataset.id,
                            "volume_id": volume_id,
                            "volume_name": volume_name,
                            "ann_json": ann_json,
                        },
                    )
                    raise e

                volume_file_path = dataset_fs.generate_item_path(volume_name)
                if download_volumes is True:
                    header = None
                    item_progress = None
                    if ds_progress is not None:
                        item_progress = tqdm_sly(
                            desc=f"Downloading '{volume_name}'",
                            total=volume_info.sizeb,
                            unit="B",
                            unit_scale=True,
                            leave=False,
                        )
                        api.video.download_path(volume_id, volume_file_path, item_progress)
                    else:
                        api.volume.download_path(volume_id, volume_file_path)
                else:
                    touch(volume_file_path)
                    header = _create_volume_header(ann)

                mask_ids = []
                mask_paths = []
                mesh_ids = []
                mesh_paths = []
                for sf in ann.spatial_figures:
                    figure_id = key_id_map.get_figure_id(sf.key())
                    if sf.geometry.name() == Mask3D.name():
                        mask_ids.append(figure_id)
                        figure_path = dataset_fs.get_mask_path(volume_name, sf)
                        mask_paths.append(figure_path)
                    if sf.geometry.name() == ClosedSurfaceMesh.name():
                        mesh_ids.append(figure_id)
                        figure_path = dataset_fs.get_interpolation_path(volume_name, sf)
                        mesh_paths.append(figure_path)
                
                figs = api.volume.figure.download(dataset.id, [volume_id], skip_geometry=True)
                figs = figs.get(volume_id, {})
                figs_ids_map = {fig.id: fig for fig in figs}
                for ann_fig in ann.figures + ann.spatial_figures:
                    fig = figs_ids_map.get(ann_fig.geometry.sly_id)
                    ann_fig.custom_data.update(fig.custom_data)

                api.volume.figure.download_stl_meshes(mesh_ids, mesh_paths)
                api.volume.figure.download_sf_geometries(mask_ids, mask_paths)

                # prepare a list of paths where converted STLs will be stored
                nrrd_paths = []
                for file in mesh_paths:
                    file = re.sub(r"\.[^.]+$", ".nrrd", file)
                    file = change_directory_at_index(file, "mask", -3)  # change destination folder
                    nrrd_paths.append(file)

                stl_converter.to_nrrd(mesh_paths, nrrd_paths, header=header)

                ann, meta = api.volume.annotation._update_on_transfer(
                    "download", ann, project_fs.meta, nrrd_paths
                )

                project_fs.set_meta(meta)

                dataset_fs.add_item_file(
                    volume_name,
                    volume_file_path,
                    ann=ann,
                    _validate_item=False,
                )

                if progress_cb is not None:
                    progress_cb(1)

            if log_progress:
                ds_progress(len(batch))

    project_fs.set_key_id_map(key_id_map)


def load_figure_data(
    api: Api, volume_file_path: str, spatial_figure: VolumeFigure, key_id_map: KeyIdMap
):
    """
    Load data into figure geometry.

    :param api: Supervisely API address and token.
    :type api: Api
    :param volume_file_path: Path to Volume file location
    :type volume_file_path: str
    :param spatial_figure: Spatial figure
    :type spatial_figure: VolumeFigure object
    :param key_id_map: Mapped keys and IDs
    :type key_id_map: KeyIdMap object
    """
    figure_id = key_id_map.get_figure_id(spatial_figure.key())
    figure_path = "{}_mask3d/".format(volume_file_path[:-5]) + f"{figure_id}.nrrd"
    api.volume.figure.download_stl_meshes([figure_id], [figure_path])
    Mask3D.from_file(spatial_figure, figure_path)


# TODO: add methods to convert to 3d masks


def upload_volume_project(
    dir: str,
    api: Api,
    workspace_id: int,
    project_name: Optional[str] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> Tuple[int, str]:
    project_fs = VolumeProject.read_single(dir)
    if project_name is None:
        project_name = project_fs.name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project = api.project.create(workspace_id, project_name, ProjectType.VOLUMES)
    api.project.update_meta(project.id, project_fs.meta.to_json())

    if progress_cb is not None:
        log_progress = False

    item_id_dct, anns_paths_dct, interpolation_dirs_dct, mask_dirs_dct = {}, {}, {}, {}

    for dataset_fs in project_fs.datasets:
        dataset_fs: VolumeDataset
        dataset = api.dataset.create(project.id, dataset_fs.name)

        names, item_paths, ann_paths, mask_dirs, interpolation_dirs = [], [], [], [], []
        for item_name in dataset_fs:
            img_path, ann_path = dataset_fs.get_item_paths(item_name)
            names.append(item_name)
            item_paths.append(img_path)
            ann_paths.append(ann_path)
            interpolation_dirs.append(dataset_fs.get_interpolation_dir(item_name))
            mask_dirs.append(dataset_fs.get_mask_dir(item_name))

        ds_progress = progress_cb
        if log_progress is True:
            ds_progress = tqdm_sly(
                desc="Uploading volumes to {!r}".format(dataset.name),
                total=len(item_paths),
                position=0,
            )

        item_infos = api.volume.upload_nrrd_series_paths(
            dataset.id, names, item_paths, ds_progress, log_progress
        )
        volume_ids = [item_info.id for item_info in item_infos]

        anns_progress = None
        if log_progress is True or progress_cb is not None:
            anns_progress = tqdm_sly(
                desc="Uploading annotations to {!r}".format(dataset.name),
                total=len(volume_ids),
                leave=False,
            )
        api.volume.annotation.upload_paths(
            volume_ids,
            ann_paths,
            project_fs.meta,
            interpolation_dirs,
            anns_progress,
            mask_dirs,
        )

    return project.id, project.name


def _create_volume_header(ann: VolumeAnnotation) -> Dict:
    """
    Create volume header to use in STL converter when downloading project without volumes.

    :param ann: VolumeAnnotation object
    :type ann: VolumeAnnotation
    :return: header with Volume meta parameters
    :rtype: Dict
    """
    header = {}
    header["sizes"] = numpy.array([value for _, value in ann.volume_meta["dimensionsIJK"].items()])
    world_matrix = ann.volume_meta["IJK2WorldMatrix"]
    header["space directions"] = numpy.array(
        [world_matrix[i : i + 3] for i in range(0, len(world_matrix) - 4, 4)]
    )
    header["space origin"] = numpy.array(
        [world_matrix[i + 3] for i in range(0, len(world_matrix) - 4, 4)]
    )
    if ann.volume_meta["ACS"] == "RAS":
        header["space"] = "right-anterior-superior"
    elif ann.volume_meta["ACS"] == "LAS":
        header["space"] = "left-anterior-superior"
    elif ann.volume_meta["ACS"] == "LPS":
        header["space"] = "left-posterior-superior"
    return header



from typing import Callable, List, Optional, Union

from supervisely import ObjClass, ObjClassCollection
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import NotificationBox, Text, Widget
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh
from supervisely.geometry.cuboid_2d import Cuboid2d
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.mask_3d import Mask3D
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.oriented_bbox import OrientedBBox
from supervisely.geometry.point import Point
from supervisely.geometry.point_3d import Point3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.color import generate_rgb

type_to_shape_text = {
    AnyGeometry: "any shape",
    Rectangle: "rectangle",
    Polygon: "polygon",
    AlphaMask: "alpha mask",
    Bitmap: "bitmap (mask)",
    Polyline: "polyline",
    Point: "point",
    Cuboid2d: "cuboid 2d",  #
    Cuboid3d: "cuboid 3d",
    Pointcloud: "pointcloud",  # "zmdi zmdi-border-clear"
    MultichannelBitmap: "n-channel mask",  # "zmdi zmdi-collection-item"
    Point3d: "point 3d",  # "zmdi zmdi-select-all"
    GraphNodes: "keypoints",
    ClosedSurfaceMesh: "volume (3d mask)",
    Mask3D: "3d mask",
    OrientedBBox: "oriented bbox",
}

shape_text_to_type = {v: k for k, v in type_to_shape_text.items()}

# Geometry types available for creating new classes (excluding GraphNodes)
available_geometry_types = [
    {"value": "rectangle", "label": "Rectangle"},
    {"value": "polygon", "label": "Polygon"},
    {"value": "bitmap (mask)", "label": "Bitmap (mask)"},
    {"value": "polyline", "label": "Polyline"},
    {"value": "point", "label": "Point"},
    {"value": "any shape", "label": "Any shape"},
    {"value": "oriented bbox", "label": "Oriented Bounding Box"},
]


class ClassesListSelector(Widget):
    class Routes:
        CHECKBOX_CHANGED = "checkbox_cb"
        CLASS_CREATED = "class_created_cb"

    def __init__(
        self,
        classes: Optional[Union[List[ObjClass], ObjClassCollection]] = [],
        multiple: Optional[bool] = False,
        empty_notification: Optional[NotificationBox] = None,
        allow_new_classes: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        # Convert to list for internal use to allow mutations when adding new classes
        if isinstance(classes, ObjClassCollection):
            self._classes = list(classes)
        else:
            self._classes = list(classes) if classes else []
        self._multiple = multiple
        self._allow_new_classes = allow_new_classes
        self._class_created_handled = False
        if empty_notification is None:
            empty_notification = NotificationBox(
                title="No classes",
                description="No classes to select.",
            )
        self.empty_notification = empty_notification
        self._error_message = Text("", status="error", font_size=13)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        classes_list = []
        for cls in self._classes:
            shape_text = type_to_shape_text.get(cls.geometry_type)
            class_dict = {**cls.to_json(), "shape_text": shape_text.upper() if shape_text else ""}
            classes_list.append(class_dict)
        return {
            "classes": classes_list,
            "availableGeometryTypes": available_geometry_types,
        }

    def get_json_state(self):
        return {
            "selected": [False for _ in self._classes],
            "createClassDialog": {
                "visible": False,
                "className": "",
                "geometryType": "rectangle",
                "showError": False,
            },
        }

    def set(self, classes: Union[List[ObjClass], ObjClassCollection]):
        selected_classes = [cls.name for cls in self.get_selected_classes()]
        # Convert to list for internal use
        if isinstance(classes, ObjClassCollection):
            self._classes = list(classes)
        else:
            self._classes = list(classes) if classes else []
        StateJson()[self.widget_id]["selected"] = [
            cls.name in selected_classes for cls in self._classes
        ]
        self.update_data()
        StateJson().send_changes()

    def get_selected_classes(self):
        selected = StateJson()[self.widget_id]["selected"]
        return [cls for cls, is_selected in zip(self._classes, selected) if is_selected]

    def select_all(self):
        StateJson()[self.widget_id]["selected"] = [True for _ in self._classes]
        StateJson().send_changes()

    def deselect_all(self):
        StateJson()[self.widget_id]["selected"] = [False for _ in self._classes]
        StateJson().send_changes()

    def select(self, names: List[str]):
        selected = [cls.name in names for cls in self._classes]
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def deselect(self, names: List[str]):
        selected = StateJson()[self.widget_id]["selected"]
        for idx, cls in enumerate(self._classes):
            if cls.name in names:
                selected[idx] = False
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def set_multiple(self, value: bool):
        self._multiple = value

    def get_all_classes(self):
        return self._classes

    def _show_error(self, message: str):
        """Show error message in the create class dialog."""
        self._error_message.text = message
        StateJson()[self.widget_id]["createClassDialog"]["showError"] = True
        StateJson().send_changes()

    def _hide_dialog(self):
        """Hide the create class dialog and reset its state."""
        state_obj = StateJson()[self.widget_id]["createClassDialog"]
        state_obj["visible"] = False
        state_obj["className"] = ""
        state_obj["showError"] = False
        StateJson().send_changes()

    def _add_new_class(self, new_class: ObjClass):
        """Add a new class to the widget and update the UI."""
        # Add to classes list
        self._classes.append(new_class)

        # Add selection state for the new class (selected by default)
        StateJson()[self.widget_id]["selected"].append(True)

        # Update data to reflect the new class in the UI
        self.update_data()
        DataJson().send_changes()
        StateJson().send_changes()

    def selection_changed(self, func):
        route_path = self.get_route_path(ClassesListSelector.Routes.CHECKBOX_CHANGED)
        server = self._sly_app.get_server()
        self._checkboxes_handled = True

        @server.post(route_path)
        def _click():
            selected = self.get_selected_classes()
            func(selected)

        return _click

    def class_created(self, func: Callable[[ObjClass], None]):
        """
        Decorator to handle new class creation event.
        The decorated function receives the newly created ObjClass.

        :param func: Function to be called when a new class is created
        :type func: Callable[[ObjClass], None]
        """
        route_path = self.get_route_path(ClassesListSelector.Routes.CLASS_CREATED)
        server = self._sly_app.get_server()
        self._class_created_handled = True

        @server.post(route_path)
        def _class_created():
            state = StateJson()[self.widget_id]["createClassDialog"]
            class_name = state["className"].strip()
            geometry_type_str = state["geometryType"]

            if not class_name:
                self._show_error("Class name cannot be empty")
                return

            if any(cls.name == class_name for cls in self._classes):
                self._show_error(f"Class '{class_name}' already exists")
                return

            geometry_type = shape_text_to_type.get(geometry_type_str)
            if geometry_type is None:
                self._show_error("Invalid geometry type")
                return

            # Generate color for the new class
            existing_colors = [cls.color for cls in self._classes]
            new_color = generate_rgb(existing_colors)

            # Create new class
            new_class = ObjClass(name=class_name, geometry_type=geometry_type, color=new_color)

            self._add_new_class(new_class)
            self._hide_dialog()

            func(new_class)

        return _class_created



import functools
import time

from supervisely.sly_logger import logger


def timeit(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()  # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3
        fname = func.__name__
        if hasattr(func, "__qualname__"):
            fname = func.__qualname__
        logger.debug(f"TIME {fname!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


def timeit_with_result(func):
    """Measures execution time and stores it in function's 'elapsed' attribute."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        try:
            result = func(*args, **kwargs)
            return result
        finally:
            wrapper.elapsed = time.perf_counter() - start_time
            logger.debug(
                f"Function '{func.__name__}' finished in {wrapper.elapsed:.2f} seconds (≈ {wrapper.elapsed/60:.2f} minutes)"
            )

    wrapper.elapsed = None  # Initial value before first call
    return wrapper


def update_fields(func):
    """Update state field after executing function"""

    @functools.wraps(func)
    def wrapper_updater(*args, **kwargs):
        kwargs["fields_to_update"] = {}
        exception = None

        try:
            value = func(*args, **kwargs)
        except Exception as ex:
            value = None
            exception = ex

        user_api = kwargs.get("api", None)
        app_task_id = kwargs.get("task_id", None)

        if user_api and app_task_id and len(kwargs["fields_to_update"]) > 0:
            user_api.task.set_fields_from_dict(app_task_id, kwargs["fields_to_update"])

        if exception:
            raise exception

        return value

    return wrapper_updater



import os
from collections import defaultdict
from typing import Dict, Union, Optional

from supervisely import ProjectMeta, generate_free_name, is_development, logger
from supervisely.api.api import Api, ApiContext
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.imaging.image import SUPPORTED_IMG_EXTS
from supervisely.io.fs import get_file_ext, list_files
from supervisely.project.project_settings import LabelingInterface


class MultiViewImageConverter(ImageConverter):

    def __init__(
            self,
            input_data: str,
            labeling_interface: Optional[Union[LabelingInterface, str]],
            upload_as_links: bool,
            remote_files_map: Optional[Dict[str, str]] = None,
    ):
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self._supports_links = True
        self._force_shape_for_links = self.upload_as_links

    def __str__(self):
        return AvailableImageConverters.MULTI_VIEW

    def validate_labeling_interface(self) -> bool:
        """Only multi_view labeling interface can be used to group images on single screen."""
        return self._labeling_interface == LabelingInterface.MULTIVIEW

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")
        group_map = self._find_image_directories()
        if not group_map:
            logger.debug(f"Input data does not match {str(self)} format.")
            return False
        else:
            self._group_map = group_map
            logger.debug(f"Found multi-view images in {self._input_data}.")
            return True

    def _find_image_directories(self) -> Union[Dict[str, list], None]:
        group_map = defaultdict(list)
        ann_exts = [".json", ".xml", ".txt"]
        for root, _, files in os.walk(self._input_data):
            for file in files:
                if get_file_ext(file) in ann_exts:
                    return None
                if get_file_ext(file) in SUPPORTED_IMG_EXTS:
                    group_map[root].append(os.path.join(root, file))
            
        return group_map

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Upload converted data to Supervisely"""
        dataset = api.dataset.get_info_by_id(dataset_id)
        project_id = dataset.project_id
        api.project.set_multiview_settings(project_id)

        meta_json = api.project.get_meta(project_id, with_settings=True)
        meta = ProjectMeta.from_json(meta_json)

        existing_names = set([info.name for info in api.image.get_list(dataset.id)])
        items_count = sum(len(images) for images in self._group_map.values())
        if log_progress:
            progress, progress_cb = self.get_progress(items_count, "Uploading images...")
        else:
            progress_cb = None

        for group_path, image_group in self._group_map.items():
            logger.info(f"Found files in {group_path}.")

            group_name = os.path.basename(group_path.rstrip("/"))
            images = []
            for image in image_group:
                name = os.path.basename(image)
                new_name = generate_free_name(
                    existing_names, name, with_ext=True, extend_used_names=True
                )
                if new_name != name:
                    logger.warn(f"Image '{name}' already exists. Renamed to '{new_name}'.")
                    os.rename(image, os.path.join(group_path, new_name))
                    image = os.path.join(group_path, new_name)
                if self._upload_as_links:
                    image = self.remote_files_map.get(image, image)
                images.append(image)

            with ApiContext(
                api=api, project_id=project_id, dataset_id=dataset_id, project_meta=meta
            ):
                kwarg = {"links": images} if self.upload_as_links else {"paths": images}
                api.image.upload_multiview_images(
                    dataset.id, group_name, **kwarg, progress_cb=progress_cb
                )

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset '{dataset.name}' has been successfully uploaded.")



from __future__ import annotations

import asyncio
import contextlib
import copy
import enum
import json
import os
import queue
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union

import jsonpatch
from fastapi import Request

from supervisely._utils import is_production
from supervisely.api.api import Api
import supervisely.app.fastapi.multi_user as multi_user
from supervisely.app.fastapi import run_sync
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.singleton import Singleton
from supervisely.io import env as sly_env
from supervisely.io.fs import dir_exists, mkdir
from supervisely.sly_logger import logger

_pool = ThreadPoolExecutor()


@contextlib.asynccontextmanager
async def async_lock(lock: threading.Lock):
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(_pool, lock.acquire)
    try:
        yield  # the lock is held
    finally:
        lock.release()


class Field(str, enum.Enum):
    STATE = "state"
    DATA = "data"
    CONTEXT = "context"


def get_data_dir():
    dir = None

    task_id = sly_env.task_id(raise_not_found=False)
    if task_id is not None:
        dir = f"/sessions/{task_id}"

    # order matters
    keys = ["SLY_APP_DATA_DIR", "DEBUG_APP_DIR"]

    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            dir = value
            logger.debug(f"Load dir from env {key}={value}")
            break
    if dir is None:
        raise ValueError(f"One of the env variables have to be defined: {[*keys, 'TASK_ID']}")

    if dir_exists(dir) is False:
        logger.info(f"App data directory {dir} doesn't exist. Will be made automatically.")
        mkdir(dir)
    return dir


def get_synced_data_dir():
    dir = "/sly-app-data"

    # order matters
    keys = ["SLY_APP_DATA_DIR", "DEBUG_APP_DIR"]

    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            dir = value
            logger.debug(f"Load dir from env {key}={value}")
            break

    if dir_exists(dir) is False:
        logger.info(f"Synced app data directory {dir} doesn't exist. Will be made automatically.")
        mkdir(dir)
    return dir


class _PatchableJson(dict):
    def __init__(self, field: Field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ws = WebsocketManager()
        self._last = copy.deepcopy(dict(self))
        self._lock = threading.Lock()
        self._field = field.value

    def get_changes(self, patch=None):
        if patch is None:
            patch = self._get_patch()
        return {self._field: json.loads(patch.to_string())}

    def _get_patch(self):
        patch = jsonpatch.JsonPatch.from_diff(self._last, self)
        return patch

    async def _apply_patch(self, patch):
        async with async_lock(self._lock):
            patch.apply(self._last, in_place=True)
            self._last = copy.deepcopy(self._last)

    async def synchronize_changes(self, user_id: Optional[Union[int, str]] = None):
        patch = self._get_patch()
        if user_id is not None:
            async with multi_user.async_session_context(user_id):
                await self._apply_patch(patch)
                await self._ws.broadcast(
                    self.get_changes(patch), user_id=user_id
                )
        else:
            await self._apply_patch(patch)
            await self._ws.broadcast(self.get_changes(patch), user_id=user_id)

    async def send_changes_async(self):
        user_id = None
        if sly_env.is_multiuser_mode_enabled():
            user_id = sly_env.user_from_multiuser_app()
        await self.synchronize_changes(user_id=user_id)

    def send_changes(self):
        user_id = None
        if sly_env.is_multiuser_mode_enabled():
            user_id = sly_env.user_from_multiuser_app()
        run_sync(self.synchronize_changes(user_id=user_id))

    def raise_for_key(self, key: str):
        if key in self:
            raise KeyError(f"Key {key} already exists in {self._field}")


class StateJson(_PatchableJson, metaclass=Singleton):
    _global_lock: threading.Lock = None

    def __init__(self, *args, **kwargs):
        if StateJson._global_lock is None:
            StateJson._global_lock = threading.Lock()
        super().__init__(Field.STATE, *args, **kwargs)

    async def _apply_patch(self, patch):
        await super()._apply_patch(patch)
        # @TODO: _replace_global to patching for optimization
        await StateJson._replace_global(dict(self))

    @classmethod
    async def from_request(cls, request: Request, local: bool = True) -> StateJson:
        if "application/json" not in request.headers.get("Content-Type", ""):
            return None
        content = await request.json()

        if content.get("context", {}).get("outside_request", False) is True:
            return None
        # TODO: should we always replace STATE with {}?
        d = content.get(Field.STATE, {})
        await cls._replace_global(d)

        return cls(d, __local__=local)

    @classmethod
    async def _replace_global(cls, d: dict):
        # pylint: disable=not-async-context-manager
        async with async_lock(cls._global_lock):
            global_state = cls()
            # !!! May cause problems with some apps !!!
            # global_state.clear()
            global_state.update(copy.deepcopy(d))
            global_state._last = copy.deepcopy(d)
            ContentOrigin().update(state=copy.deepcopy(d))


class DataJson(_PatchableJson, metaclass=Singleton):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.DATA, *args, **kwargs)

    async def _apply_patch(self, patch):
        async with async_lock(self._lock):
            patch.apply(self._last, in_place=True)
            self._last = copy.deepcopy(self._last)
            ContentOrigin().update(data_patch=copy.deepcopy(patch))


class ContentOrigin(metaclass=Singleton):
    def __init__(self):
        self._SLEEP_TIME = sly_env.content_origin_update_interval()
        self._data_patch_queue = queue.Queue()
        self._last_sent_data = {}
        self._state_queue = queue.Queue()
        self._stop = threading.Event()
        self._loop_thread = threading.Thread(
            target=self._update_content_loop, name="ContentOrigin._update_content_loop"
        )

    def start(self):
        if not self._loop_thread.is_alive():
            self._loop_thread.start()
            self._stop.clear()
            DataJson().send_changes()
            StateJson().send_changes()

    def stop(self):
        self._stop.set()

    def update(self, data_patch=None, state=None):
        if is_production():
            if data_patch is not None:
                self._data_patch_queue.put(data_patch)
            if state is not None:
                self._state_queue.put(state)

    def _send(self, data_patch: jsonpatch.JsonPatch, state: dict):
        task_id = sly_env.task_id()
        api = Api()
        api.task._update_app_content(task_id, data_patch=list(data_patch), state=state)

    def _update_content_loop(self):
        failed_patch = None
        while True:
            last_state = None
            state_count = 0
            patches = []
            while not self._data_patch_queue.empty():
                patches.append(self._data_patch_queue.get())
            while not self._state_queue.empty():
                last_state = self._state_queue.get()
                state_count += 1

            if patches or last_state is not None:
                try:
                    merged_patch = None
                    data = copy.deepcopy(self._last_sent_data)
                    for patch in [failed_patch, *patches]:
                        if patch is None:
                            continue
                        patch.apply(data, in_place=True)
                    merged_patch = jsonpatch.JsonPatch.from_diff(self._last_sent_data, data)
                    self._send(data_patch=merged_patch, state=last_state)
                    self._last_sent_data = copy.deepcopy(data)
                    failed_patch = None
                except Exception as exc:
                    failed_patch = merged_patch
                    logger.error(
                        traceback.format_exc(),
                        exc_info=True,
                        extra={"exc_str": str(exc)},
                    )
                finally:
                    for _ in range(len(patches)):
                        self._data_patch_queue.task_done()
                    for _ in range(state_count):
                        self._state_queue.task_done()

            elif self._stop.is_set():
                return

            time.sleep(self._SLEEP_TIME)



from typing import Callable, List, Optional, Union

from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Text, Widget
from supervisely.imaging.color import generate_rgb

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


# Available value types for tag creation
available_value_types = [
    {"value": TagValueType.NONE, "label": "None"},
    {"value": TagValueType.ANY_STRING, "label": "Any string"},
    {"value": TagValueType.ANY_NUMBER, "label": "Any number"},
    {"value": TagValueType.ONEOF_STRING, "label": "One of"},
]


class SelectTag(Widget):
    """
    SelectTag is a compact dropdown widget for selecting tag metadata with an option to create
    new tags on the fly.

    :param tags: Initial list of TagMeta instances
    :type tags: Optional[Union[List[TagMeta], TagMetaCollection]]
    :param filterable: Enable search/filter functionality in dropdown
    :type filterable: Optional[bool]
    :param placeholder: Placeholder text when no tag is selected
    :type placeholder: Optional[str]
    :param show_add_new_tag: Show "Add new tag" option at the end of the list
    :type show_add_new_tag: Optional[bool]
    :param size: Size of the select dropdown
    :type size: Optional[Literal["large", "small", "mini"]]
    :param multiple: Enable multiple selection
    :type multiple: bool
    :param widget_id: Unique widget identifier
    :type widget_id: Optional[str]

    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.app.widgets import SelectTag

        # Create some initial tags
        tag_weather = sly.TagMeta('weather', sly.TagValueType.ANY_STRING)
        tag_count = sly.TagMeta('count', sly.TagValueType.ANY_NUMBER)

        colors = ["red", "green", "blue"]
        tag_color = sly.TagMeta('color', sly.TagValueType.ONEOF_STRING, possible_values=colors)

        # Create SelectTag widget
        select_tag = SelectTag(
            tags=[tag_weather, tag_count, tag_color],
            filterable=True,
            show_add_new_tag=True
        )

        # Handle selection changes
        @select_tag.value_changed
        def on_tag_selected(tag_name):
            print(f"Selected tag: {tag_name}")
            selected_tag = select_tag.get_selected_tag()
            print(f"Tag object: {selected_tag}")

        # Handle new tag creation
        @select_tag.tag_created
        def on_tag_created(new_tag: sly.TagMeta):
            print(f"New tag created: {new_tag.name}")
    """

    class Routes:
        VALUE_CHANGED = "value_changed"
        TAG_CREATED = "tag_created_cb"

    def __init__(
        self,
        tags: Optional[Union[List[TagMeta], TagMetaCollection]] = [],
        filterable: Optional[bool] = True,
        placeholder: Optional[str] = "Select tag",
        show_add_new_tag: Optional[bool] = True,
        size: Optional[Literal["large", "small", "mini"]] = None,
        multiple: bool = False,
        widget_id: Optional[str] = None,
    ):
        # Convert to list for internal use to allow mutations when adding new tags
        if isinstance(tags, TagMetaCollection):
            self._tags = list(tags)
        else:
            self._tags = list(tags) if tags else []

        self._filterable = filterable
        self._placeholder = placeholder
        self._show_add_new_tag = show_add_new_tag
        self._size = size
        self._multiple = multiple

        self._changes_handled = False
        self._tag_created_callback = None

        # Store error message widget
        self._error_message = Text("", status="error", font_size=13)

        # Initialize parent Widget
        super().__init__(widget_id=widget_id, file_path=__file__)

        # Register tag_created route if show_add_new_tag is enabled
        if self._show_add_new_tag:
            self._register_tag_created_route()

    def get_json_data(self):
        """Build JSON data for the widget."""
        # Build items list with tag info
        items = []
        for tag in self._tags:
            value_type_text = tag.value_type.replace("_", " ").upper()

            items.append(
                {
                    "value": tag.name,
                    "label": tag.name,
                    "color": tag.color,
                    "valueType": value_type_text if value_type_text else "",
                }
            )

        return {
            "items": items,
            "placeholder": self._placeholder,
            "filterable": self._filterable,
            "multiple": self._multiple,
            "size": self._size,
            "showAddNewTag": self._show_add_new_tag,
            "availableValueTypes": available_value_types,
        }

    def get_json_state(self):
        """Build JSON state for the widget."""
        # Set initial value based on multiple mode
        if self._multiple:
            initial_value = []
        else:
            initial_value = self._tags[0].name if self._tags else None

        return {
            "value": initial_value,
            "createTagDialog": {
                "visible": False,
                "tagName": "",
                "valueType": TagValueType.NONE,
                "possibleValues": "",
                "showError": False,
                "showPossibleValues": False,
            },
        }

    def get_value(self) -> Union[str, List[str], None]:
        """Get the currently selected tag name(s)."""
        return StateJson()[self.widget_id]["value"]

    def get_selected_tag(self) -> Union[TagMeta, List[TagMeta], None]:
        """Get the currently selected TagMeta object(s)."""
        value = self.get_value()
        if value is None:
            return None

        if self._multiple:
            if not isinstance(value, list):
                return []
            result = []
            for tag_name in value:
                for tag in self._tags:
                    if tag.name == tag_name:
                        result.append(tag)
                        break
            return result
        else:
            for tag in self._tags:
                if tag.name == value:
                    return tag
            return None

    def set_value(self, tag_name: Union[str, List[str]]):
        """Set the selected tag by name."""
        StateJson()[self.widget_id]["value"] = tag_name
        StateJson().send_changes()

    def get_all_tags(self) -> List[TagMeta]:
        """Get all available tags."""
        return self._tags.copy()

    def set(self, tags: Union[List[TagMeta], TagMetaCollection]):
        """Update the list of available tags."""
        # Convert to list for internal use
        if isinstance(tags, TagMetaCollection):
            self._tags = list(tags)
        else:
            self._tags = list(tags) if tags else []

        # Update data
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

        # Reset value if current selection is not in new tags
        current_value = StateJson()[self.widget_id]["value"]
        if current_value:
            if self._multiple:
                if isinstance(current_value, list):
                    # Keep only valid selections
                    valid = [v for v in current_value if any(tag.name == v for tag in self._tags)]
                    if valid != current_value:
                        StateJson()[self.widget_id]["value"] = valid
                        StateJson().send_changes()
            else:
                if not any(tag.name == current_value for tag in self._tags):
                    StateJson()[self.widget_id]["value"] = (
                        self._tags[0].name if self._tags else None
                    )
                    StateJson().send_changes()

    def _show_error(self, message: str):
        """Show error message in the create tag dialog."""
        self._error_message.text = message
        StateJson()[self.widget_id]["createTagDialog"]["showError"] = True
        StateJson().send_changes()

    def _hide_dialog(self):
        """Hide the create tag dialog and reset its state."""
        state_obj = StateJson()[self.widget_id]["createTagDialog"]
        state_obj["visible"] = False
        state_obj["tagName"] = ""
        state_obj["possibleValues"] = ""
        state_obj["showError"] = False
        state_obj["showPossibleValues"] = False
        StateJson().send_changes()

    def _add_new_tag(self, new_tag: TagMeta):
        """Add a new tag to the widget and update the UI."""
        # Add to tags list
        self._tags.append(new_tag)

        # Update data
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

        # Set the new tag as selected
        if self._multiple:
            current = StateJson()[self.widget_id]["value"]
            if isinstance(current, list):
                current.append(new_tag.name)
            else:
                StateJson()[self.widget_id]["value"] = [new_tag.name]
        else:
            StateJson()[self.widget_id]["value"] = new_tag.name
        StateJson().send_changes()

    def value_changed(self, func: Callable[[Union[TagMeta, List[TagMeta]]], None]):
        """
        Decorator to handle value change event.
        The decorated function receives the selected TagMeta (or list of TagMeta if multiple=True).

        :param func: Function to be called when selection changes
        :type func: Callable[[Union[TagMeta, List[TagMeta]]], None]
        """
        route_path = self.get_route_path(SelectTag.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            selected = self.get_selected_tag()
            if selected is not None:
                func(selected)

        return _value_changed

    def _register_tag_created_route(self):
        """Register the tag_created route."""
        route_path = self.get_route_path(SelectTag.Routes.TAG_CREATED)
        server = self._sly_app.get_server()

        @server.post(route_path)
        def _tag_created():
            state = StateJson()[self.widget_id]["createTagDialog"]
            tag_name = state["tagName"].strip()
            value_type = state["valueType"]
            possible_values_str = state["possibleValues"].strip()

            # Validate tag name
            if not tag_name:
                self._show_error("Tag name cannot be empty")
                return

            # Check if tag with this name already exists
            if any(tag.name == tag_name for tag in self._tags):
                self._show_error(f"Tag '{tag_name}' already exists")
                return

            # Parse possible values for ONEOF_STRING
            possible_values = None
            if value_type == TagValueType.ONEOF_STRING:
                if not possible_values_str:
                    self._show_error("Possible values are required for 'One of' type")
                    return
                # Split by comma and strip whitespace
                possible_values = [v.strip() for v in possible_values_str.split(",") if v.strip()]
                if len(possible_values) == 0:
                    self._show_error("At least one possible value is required")
                    return

            # Generate color for the new tag
            existing_colors = [tag.color for tag in self._tags]
            new_color = generate_rgb(existing_colors)

            # Create new tag
            try:
                new_tag = TagMeta(
                    name=tag_name,
                    value_type=value_type,
                    possible_values=possible_values,
                    color=new_color,
                )
            except Exception as e:
                self._show_error(f"Error creating tag: {str(e)}")
                return

            # Add to widget
            self._add_new_tag(new_tag)

            # Hide dialog
            self._hide_dialog()

            # Call user's callback if set
            if self._tag_created_callback:
                self._tag_created_callback(new_tag)

    def tag_created(self, func: Callable[[TagMeta], None]):
        """
        Decorator to handle new tag creation event.
        The decorated function receives the newly created TagMeta.

        :param func: Function to be called when a new tag is created
        :type func: Callable[[TagMeta], None]
        """
        self._tag_created_callback = func
        return func



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Tuple, Type

import supervisely.io.env as env
from supervisely.api.agent_api import AgentInfo
from supervisely.api.api import Api
from supervisely.io.fs import get_file_name_with_ext
from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts
from supervisely.nn.experiments import ExperimentInfo


def find_agent(api: "Api", team_id: int = None, public=True, gpu=True) -> int:
    """
    Find an agent in Supervisely with most available GPU memory.

    :param api: Supervisely API client.
    :type api: :class:`~supervisely.api.api.Api`
    :param team_id: Team ID. If not provided, will be taken from the current context.
    :type team_id: int, optional
    :param public: If True, can find a public agent.
    :type public: bool
    :param gpu: If True, find an agent with GPU.
    :type gpu: bool
    :return: Agent ID.
    :rtype: int
    :raises ValueError: If no available agents found.
    """
    if team_id is None:
        team_id = env.team_id()
    agents = api.agent.get_list_available(team_id, show_public=public, has_gpu=gpu)
    if len(agents) == 0:
        raise ValueError("No available agents found.")
    agent_id_memory_map = {}
    kubernetes_agents = []
    for agent in agents:
        if agent.type == "sly_agent":
            # No multi-gpu support, always take the first one
            agent_id_memory_map[agent.id] = agent.gpu_info["device_memory"][0]["available"]
        elif agent.type == "kubernetes":
            kubernetes_agents.append(agent.id)
    if len(agent_id_memory_map) > 0:
        return max(agent_id_memory_map, key=agent_id_memory_map.get)
    if len(kubernetes_agents) > 0:
        return kubernetes_agents[0]


def get_framework_by_path(path: str) -> Type[BaseTrainArtifacts]:
    from supervisely.nn.artifacts import (
        RITM,
        RTDETR,
        Detectron2,
        MMClassification,
        MMDetection,
        MMDetection3,
        MMPretrain,
        MMSegmentation,
        UNet,
        YOLOv5,
        YOLOv5v2,
        YOLOv8,
    )

    path_obj = Path(path)
    if len(path_obj.parts) < 2:
        raise ValueError(f"Incorrect checkpoint path: '{path}'")
    parent = path_obj.parts[1]
    frameworks = {
        "/detectron2": Detectron2,
        "/mmclassification": MMClassification,
        "/mmclassification-v2": MMPretrain,
        "/mmdetection": MMDetection,
        "/mmdetection-3": MMDetection3,
        "/mmsegmentation": MMSegmentation,
        "/RITM_training": RITM,
        "/RT-DETR": RTDETR,
        "/unet": UNet,
        "/yolov5_train": YOLOv5,
        "/yolov5_2.0_train": YOLOv5v2,
        "/yolov8_train": YOLOv8,
    }
    if f"/{parent}" in frameworks:
        return frameworks[f"/{parent}"]


def get_artifacts_dir_and_checkpoint_name(model: str) -> Tuple[str, str]:
    if not model.startswith("/"):
        raise ValueError(f"Path must start with '/'")

    if model.startswith("/experiments"):
        if model.endswith(".pt") or model.endswith(".pth"):
            try:
                artifacts_dir, checkpoint_name = model.split("/checkpoints/")
                return artifacts_dir, checkpoint_name
            except:
                raise ValueError(
                    "Bad format of checkpoint path. Expected format: '/artifacts_dir/checkpoints/checkpoint_name'"
                )
        elif model.endswith(".onnx") or model.endswith(".engine"):
            try:
                artifacts_dir, checkpoint_name = model.split("/export/")
                return artifacts_dir, checkpoint_name
            except:
                raise ValueError(
                    "Bad format of checkpoint path. Expected format: '/artifacts_dir/export/checkpoint_name'"
                )
        else:
            raise ValueError(f"Unknown model format: '{get_file_name_with_ext(model)}'")

    framework_cls = get_framework_by_path(model)
    if framework_cls is None:
        raise ValueError(f"Unknown path: '{model}'")

    team_id = env.team_id()
    framework = framework_cls(team_id)
    checkpoint_name = get_file_name_with_ext(model)
    checkpoints_dir = model.replace(checkpoint_name, "")
    if framework.weights_folder is not None:
        artifacts_dir = checkpoints_dir.replace(framework.weights_folder, "")
    else:
        artifacts_dir = checkpoints_dir
    return artifacts_dir, checkpoint_name

def find_team_by_path(api: "Api", path: str, team_id: int = None, raise_not_found=True):
    if team_id is not None:
        if api.file.exists(team_id, path) or api.file.dir_exists(
            team_id, path, recursive=False
        ):
            return team_id
        elif raise_not_found:
            raise ValueError(f"Checkpoint '{path}' not found in team provided team")
        else:
            return None
    team_id = env.team_id(raise_not_found=False)
    if team_id is not None:
        if api.file.exists(team_id, path) or api.file.dir_exists(
            team_id, path, recursive=False
        ):
            return team_id
    teams = api.team.get_list()
    team_id = None
    for team in teams:
        if api.file.exists(team.id, path):
            if team_id is not None:
                raise ValueError("Multiple teams have the same checkpoint")
            team_id = team.id
    if team_id is None:
        if raise_not_found:
            raise ValueError("Checkpoint not found")
        else:
            return None
    return team_id

def find_apps_by_framework(api: "Api", framework: str, categories: List[str] = None):
    if categories is None:
        categories = []
    modules = api.app.get_list_ecosystem_modules(
        categories=[*categories, f"framework:{framework}"], categories_operation="and"
    )
    return modules

def run_train_app(
    api: "Api",
    agent_id: int,
    module_id: int,
    workspace_id: int,
    app_state: dict,
    timeout: int = 100,
    **kwargs
):
    """
    Run a training app.

    :param api: Supervisely API client.
    :type api: :class:`~supervisely.api.api.Api`
    :param agent_id: Agent ID where the app task will run.
    :type agent_id: int
    :param module_id: Module ID of the training app.
    :type module_id: int
    :param workspace_id: Workspace ID where the app task will run.
    :type workspace_id: int
    :param app_state: App state to run the training app. Must include key state with app state inside: 'state': app_state"
    :type app_state: dict
    :param timeout: Timeout in seconds.
    :type timeout: int
    :return: Task information.
    :rtype: dict
    """

    _attempt_delay_sec = 1
    _attempts = timeout // _attempt_delay_sec

    task_info = api.task.start(
        agent_id=agent_id,
        module_id=module_id,
        workspace_id=workspace_id,
        params=app_state,
        **kwargs
    )
    ready = api.app.wait_until_ready_for_api_calls(
        task_info["id"], _attempts, _attempt_delay_sec
    )
    if not ready:
        raise TimeoutError(
            f"Task {task_info['id']} is not ready for API calls after {timeout} seconds."
        )
    return task_info

def get_experiment_info_by_task_id(api: "Api", task_id) -> Optional[ExperimentInfo]:
    task_info = api.task.get_info_by_id(task_id)
    experiment_data = task_info.get("meta", {}).get("output", {}).get("experiment", {}).get("data")
    if experiment_data is None:
        return None
    return ExperimentInfo(**experiment_data)



from typing import List

from supervisely import ProjectMeta, generate_free_name, logger
from supervisely.api.api import Api
from supervisely.convert.base_converter import AvailableVolumeConverters
from supervisely.convert.volume.dicom import dicom_helper as h
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.volume.volume import (
    get_extension,
    inspect_dicom_series,
    read_dicom_serie_volume,
)


class DICOMConverter(VolumeConverter):
    class Item(VolumeConverter.Item):
        """Item class for DICOM series."""

        def __init__(self, serie_id: str, item_paths: List[str], volume_meta: dict):
            item_path = item_paths[0] if len(item_paths) > 0 else None
            super().__init__(item_path, volume_meta=volume_meta)

            self._serie_id: str = serie_id
            self._item_paths: List[str] = item_paths

        @property
        def item_paths(self) -> List[str]:
            return self._item_paths

        @item_paths.setter
        def item_paths(self, paths: List[str]) -> None:
            self._item_paths = paths

        @property
        def serie_id(self) -> str:
            return self._serie_id

        @serie_id.setter
        def serie_id(self, serie_id: str) -> None:
            self._serie_id = serie_id

    def __str__(self) -> str:
        return AvailableVolumeConverters.DICOM

    @property
    def ann_ext(self) -> str:
        return None

    @property
    def key_file_ext(self) -> str:
        return None

    def validate_format(self) -> bool:
        # DICOM
        series_infos = inspect_dicom_series(root_dir=self._input_data)

        # create Items
        self._items = []
        for dicom_id, dicom_paths in series_infos.items():
            if len(dicom_paths) == 0:
                logger.warn(f"Empty serie {dicom_id}, serie will be skipped")
                continue
            item_path = dicom_paths[0]
            if get_extension(path=item_path) is None:
                logger.warn(
                    f"Can not recognize file extension {item_path}, serie will be skipped"
                )
                continue

            for dicom_path in dicom_paths:
                h.read_and_convert_to_monochrome2(dicom_path)
            _, meta = read_dicom_serie_volume(dicom_paths, anonymize=True)
            item = self.Item(
                serie_id=dicom_id, item_paths=dicom_paths, volume_meta=meta
            )
            self._items.append(item)
        self._meta = ProjectMeta()

        return self.items_count > 0

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        existing_names = set([vol.name for vol in api.volume.get_list(dataset_id)])

        for item in self._items:
            serie_id, serie_paths = item.serie_id, item.item_paths
            volume_name = f"{serie_id}.nrrd"
            volume_name = generate_free_name(
                existing_names, volume_name, with_ext=True, extend_used_names=True
            )
            api.volume.upload_dicom_serie_paths(
                dataset_id=dataset_id,
                name=volume_name,
                paths=serie_paths,
                log_progress=log_progress,
                anonymize=True,
            )

        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")



from __future__ import annotations

from typing import Dict, List, Optional

from supervisely.app import DataJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class ActivityFeed(Widget):
    """ActivityFeed is a widget that displays a vertical list of activity items with status indicators.
    Similar to a timeline or activity log showing sequential events with their current status.

    Each item can contain a custom widget as content and displays a status indicator (pending, in process, completed, failed).
    Items are automatically numbered if no number is provided.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/layouts-and-containers/activity-feed>`_
        (including screenshots and examples).

    :param items: List of ActivityFeed.Item objects to display
    :type items: Optional[List[ActivityFeed.Item]]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import ActivityFeed, Text

        # Create items with custom content
        item1 = ActivityFeed.Item(
            content=Text("Processing dataset"),
            status="completed"
        )
        item2 = ActivityFeed.Item(
            content=Text("Training model"),
            status="in_progress",
            number=2
        )
        item3 = ActivityFeed.Item(
            content=Text("Generating report"),
            status="pending"
        )

        # Create activity feed
        feed = ActivityFeed(items=[item1, item2, item3])

        # Add item during runtime
        new_item = ActivityFeed.Item(
            content=Text("Deploy model"),
            status="pending"
        )
        feed.add_item(new_item)

        # Update status by item number
        feed.set_status(2, "completed")

        # Get item status
        status = feed.get_status(2)
    """

    class Item:
        """Represents a single item in the ActivityFeed.

        :param content: Widget to display as the item content
        :type content: Widget
        :param status: Status of the item (pending, in_progress, completed, failed)
        :type status: Literal["pending", "in_progress", "completed", "failed"]
        :param number: Position number in the feed (auto-assigned if not provided)
        :type number: Optional[int]
        """

        def __init__(
            self,
            content: Widget,
            status: Literal["pending", "in_progress", "completed", "failed"] = "pending",
            number: Optional[int] = None,
        ) -> ActivityFeed.Item:
            self.content = content
            self.status = status
            self.number = number
            self._validate_status()

        def _validate_status(self):
            valid_statuses = ["pending", "in_progress", "completed", "failed"]
            if self.status not in valid_statuses:
                raise ValueError(
                    f"Invalid status '{self.status}'. Must be one of: {', '.join(valid_statuses)}"
                )

        def to_json(self):
            return {
                "number": self.number,
                "status": self.status,
            }

    def __init__(
        self,
        items: Optional[List[ActivityFeed.Item]] = None,
        widget_id: Optional[str] = None,
    ):
        self._items = items if items is not None else []
        self._auto_assign_numbers()
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _auto_assign_numbers(self):
        """Automatically assign numbers to items that don't have them."""
        next_number = 1
        for item in self._items:
            if item.number is None:
                item.number = next_number
            next_number = max(next_number, item.number) + 1

    def get_json_data(self) -> Dict:
        """Returns dictionary with widget data.

        :return: Dictionary with items data
        :rtype: Dict
        """
        return {
            "items": [item.to_json() for item in self._items],
        }

    def get_json_state(self) -> Dict:
        """Returns dictionary with widget state (empty for this widget).

        :return: Empty dictionary
        :rtype: Dict
        """
        return {}

    def add_item(
        self,
        item: Optional[ActivityFeed.Item] = None,
        content: Optional[Widget] = None,
        status: Literal["pending", "in_progress", "completed", "failed"] = "pending",
        number: Optional[int] = None,
    ) -> None:
        """Add a new item to the activity feed.

        You can either pass an ActivityFeed.Item object or provide content and status separately.

        :param item: ActivityFeed.Item to add
        :type item: Optional[ActivityFeed.Item]
        :param content: Widget content (used if item is not provided)
        :type content: Optional[Widget]
        :param status: Status of the item (used if item is not provided)
        :type status: Literal["pending", "in_progress", "completed", "failed"]
        :param number: Position number (auto-assigned if not provided)
        :type number: Optional[int]
        """
        if item is None:
            if content is None:
                raise ValueError("Either 'item' or 'content' must be provided")
            item = ActivityFeed.Item(content=content, status=status, number=number)

        if item.number is None:
            # Auto-assign number
            if self._items:
                item.number = max(i.number for i in self._items) + 1
            else:
                item.number = 1

        self._items.append(item)
        self.update_data()
        DataJson().send_changes()

    def remove_item(self, number: int) -> None:
        """Remove an item from the activity feed by its number.

        :param number: Number of the item to remove. Starts from 1.
        :type number: int
        """
        self._items = [item for item in self._items if item.number != number]
        self.update_data()
        DataJson().send_changes()

    def set_status(
        self,
        number: int,
        status: Literal["pending", "in_progress", "completed", "failed"],
    ) -> None:
        """Update the status of an item by its number.

        :param number: Number of the item to update. Starts from 1.
        :type number: int
        :param status: New status for the item
        :type status: Literal["pending", "in_progress", "completed", "failed"]
        """
        for i, item in enumerate(self._items):
            if item.number == number:
                item.status = status
                item._validate_status()
                DataJson()[self.widget_id]["items"][i]["status"] = status
                DataJson().send_changes()
                return
        raise ValueError(f"Item with number {number} not found")

    def get_status(self, number: int) -> str:
        """Get the status of an item by its number.

        :param number: Number of the item. Starts from 1.
        :type number: int
        :return: Status of the item
        :rtype: str
        """
        for item in self._items:
            if item.number == number:
                return item.status
        raise ValueError(f"Item with number {number} not found")

    def get_items(self) -> List[ActivityFeed.Item]:
        """Get all items in the activity feed.

        :return: List of all items
        :rtype: List[ActivityFeed.Item]
        """
        return self._items

    def clear(self) -> None:
        """Remove all items from the activity feed."""
        self._items = []
        self.update_data()
        DataJson().send_changes()

    def set_items(self, items: List[ActivityFeed.Item]) -> None:
        """Replace all items in the activity feed.

        :param items: New list of items
        :type items: List[ActivityFeed.Item]
        """
        self._items = items
        self._auto_assign_numbers()
        self.update_data()
        DataJson().send_changes()



# coding: utf-8
"""create or manipulate already existing Entities Collection in Supervisely"""

# docs
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Union

import requests

from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    RemoveableModuleApi,
    UpdateableModule,
)
from supervisely.sly_logger import logger

if TYPE_CHECKING:
    from supervisely.api.api import Api
    from supervisely.api.image_api import ImageInfo


class CollectionType:
    DEFAULT = "default"
    AI_SEARCH = "aiSearch"
    ALL = "all"


class CollectionTypeFilter:
    AI_SEARCH = "entities_ai_search_collection"
    DEFAULT = "entities_collection"


class AiSearchThresholdDirection:
    ABOVE = "above"
    BELOW = "below"


@dataclass
class CollectionItem:
    """
    Collection item with meta information.

    :param entity_id: Supervisely ID of the item.
    :type entity_id: int
    :param meta: Meta information about the item. Optional, defaults to None.
    :type meta: :class:`CollectionItem.Meta`, optional
    """

    @dataclass
    class Meta:
        """
        Meta information about the item.

        :param score: Score value of the item that indicates search relevance.
        :type score: float
        """

        score: Optional[float] = 0.0

        def to_json(self) -> dict:
            """
            Convert meta information to a JSON-compatible dictionary.

            :return: Dictionary with meta information.
            :rtype: dict
            """
            return {ApiField.SCORE: self.score}

        @classmethod
        def from_json(cls, data: dict) -> "CollectionItem.Meta":
            """
            Create Meta from a JSON-compatible dictionary.

            :param data: Dictionary with meta information.
            :type data: dict
            :return: Meta object.
            :rtype: CollectionItem.Meta
            """
            return cls(score=data.get(ApiField.SCORE, 0.0))

    entity_id: int
    meta: Optional[Meta] = None

    def to_json(self) -> dict:
        """
        Convert collection item to a JSON-compatible dictionary.

        :return: Dictionary with collection item data.
        :rtype: dict
        """
        result = {ApiField.ENTITY_ID: self.entity_id}
        if self.meta is not None:
            result[ApiField.META] = self.meta.to_json()
        return result

    @classmethod
    def from_json(cls, data: dict) -> "CollectionItem":
        """
        Create CollectionItem from a JSON-compatible dictionary.

        :param data: Dictionary with collection item data.
        :type data: dict
        :return: CollectionItem object.
        :rtype: CollectionItem
        """
        meta_data = data.get(ApiField.META)
        meta = cls.Meta.from_json(meta_data) if meta_data is not None else None
        return cls(entity_id=data[ApiField.ENTITY_ID], meta=meta)


class EntitiesCollectionInfo(NamedTuple):
    """
    Object with entitites collection parameters from Supervisely.

    :Example:

     .. code-block:: python

        EntitiesCollectionInfo(
            id=1,
            team_id=2,
            project_id=3,
            name="collection_name",
            created_at="2023-01-01T00:00:00Z",
            updated_at="2023-01-02T00:00:00Z",
            description="This is a collection",
            options={"key": "value"},
            type="default",
            ai_search_key="search_key"
        )
    """

    #: ID of the collection.
    id: int

    #: Name of the collection.
    name: str

    #: ID of the team.
    team_id: int

    #: ID of the project.
    project_id: int

    #: Date and time when the collection was created.
    created_at: str

    #: Date and time when the collection was last updated.
    updated_at: str

    #: Description of the collection.
    description: Optional[str] = None

    #: Additional options for the collection.
    options: Dict[str, Union[str, int, bool]] = None

    #: Type of the collection.
    type: str = CollectionType.DEFAULT

    #: AI search key for the collection.
    ai_search_key: Optional[str] = None


class EntitiesCollectionApi(UpdateableModule, RemoveableModuleApi):
    """
    API for working with Entities Collection. :class:`EntitiesCollectionApi<EntitiesCollectionApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        collection = api.entities_collection.get_list(9) # api usage example
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple EntitiesCollectionInfo information about Entities Collection.

        :Example:

         .. code-block:: python

             EntitiesCollectionInfo(
                id=2,
                name='Enitites Collections #1',
                team_id=4,
                project_id=58,
                description='',
                created_at='2020-04-08T15:10:12.618Z',
                updated_at='2020-04-08T15:10:19.833Z',
            )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.TEAM_ID,
            ApiField.PROJECT_ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.DESCRIPTION,
            ApiField.OPTIONS,
            ApiField.TYPE,
            ApiField.AI_SEARCH_KEY,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of :class:`EntitiesCollectionInfo` NamedTuple.

        :return: NamedTuple name.
        :rtype: :class:`str`
        """
        return "EntitiesCollectionInfo"

    def __init__(self, api: Api):
        ModuleApi.__init__(self, api)
        self._api = api

    def _convert_json_info(self, info: dict, skip_missing=True) -> EntitiesCollectionInfo:
        """
        Differs from the original method by using skip_missing equal to True by default.
        Also unpacks 'meta' field to top level for fields in info_sequence.
        """

        def _get_value(dict, field_name, skip_missing):
            if skip_missing is True:
                return dict.get(field_name, None)
            else:
                return dict[field_name]

        if info is None:
            return None
        else:
            field_values = []
            meta = info.get("meta", {})

            for field_name in self.info_sequence():
                if type(field_name) is str:
                    # Try to get from meta first if the field exists there
                    if field_name in meta:
                        field_values.append(meta.get(field_name))
                    else:
                        field_values.append(_get_value(info, field_name, skip_missing))
                elif type(field_name) is tuple:
                    value = None
                    for sub_name in field_name[0]:
                        if value is None:
                            value = _get_value(info, sub_name, skip_missing)
                        else:
                            value = _get_value(value, sub_name, skip_missing)
                    field_values.append(value)
                else:
                    raise RuntimeError("Can not parse field {!r}".format(field_name))
            return self.InfoType(*field_values)

    def create(
        self,
        project_id: int,
        name: str,
        description: Optional[str] = None,
        type: str = CollectionType.DEFAULT,
        ai_search_key: Optional[str] = None,
        change_name_if_conflict=False,
    ) -> EntitiesCollectionInfo:
        """
        Creates Entities Collections.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param name: Entities Collection name in Supervisely.
        :type name: str
        :param description: Entities Collection description.
        :type description: str
        :param type: Type of the collection. Defaults to "default".
        :type type: str
        :param ai_search_key: AI search key for the collection. Defaults to None.
        :type ai_search_key: Optional[str]
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name. Defaults to False.
        :type change_name_if_conflict: bool
        :return: Information about new Entities Collection
        :rtype: :class:`EntitiesCollectionInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 123
            name = "Chihuahuas"
            description = "Collection of Chihuahuas"
            type = CollectionType.AI_SEARCH
            ai_search_key = "0ed6a5256433bbe32822949d563d476a"

            new_collection = api.entities_collection.create(project_id, name, description, type, ai_search_key)
            print(new_collection)
        """

        name = self._get_effective_new_name(
            parent_id=project_id,
            name=name,
            change_name_if_conflict=change_name_if_conflict,
        )

        method = "entities-collections.add"
        data = {
            ApiField.PROJECT_ID: project_id,
            ApiField.NAME: name,
            ApiField.TYPE: type,
        }
        if description is not None:
            data[ApiField.DESCRIPTION] = description
        if ai_search_key is not None:
            if type != CollectionType.AI_SEARCH:
                raise ValueError(
                    "ai_search_key is only available for creation AI Search collection type."
                )
            if ApiField.META not in data:
                data[ApiField.META] = {}
            data[ApiField.META][ApiField.AI_SEARCH_KEY] = ai_search_key
        elif type == CollectionType.AI_SEARCH:
            raise ValueError("ai_search_key is required for AI Search collection type.")
        response = self._api.post(method, data)
        return self._convert_json_info(response.json())

    def remove(self, id: int, force: bool = False):
        """
        Remove Entites Collection with the specified ID from the Supervisely server.

        If `force` is set to True, the collection will be removed permanently.
        If `force` is set to False, the collection will be disabled instead of removed.

        :param id: Entites Collection ID in Supervisely
        :type id: int
        :param force: If True, the collection will be removed permanently. Defaults to False.
        :type force: bool
        :return: None
        """
        self._api.post(
            self._remove_api_method_name(), {ApiField.ID: id, ApiField.HARD_DELETE: force}
        )

    def get_list(
        self,
        project_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        with_meta: bool = False,
        collection_type: CollectionType = CollectionType.DEFAULT,
    ) -> List[EntitiesCollectionInfo]:
        """
        Get list of information about Entities Collection for the given project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :param filters: List of filters to apply to the request. Optional, defaults to None.
        :type filters: List[Dict[str, str]], optional
        :param with_meta: If True, includes meta information in the response. Defaults to False.
        :type with_meta: bool, optional
        :param collection_type: Type of the collection.
                    Defaults to CollectionType.DEFAULT.

                    Available types are:
                     - CollectionType.DEFAULT
                     - CollectionType.AI_SEARCH
                     - CollectionType.ALL
        :type collection_type: CollectionType
        :return: List of information about Entities Collections.
        :rtype: :class:`List[EntitiesCollectionInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collections = api.entities_collection.get_list(4)
        """
        method = "entities-collections.list"

        data = {
            ApiField.PROJECT_ID: project_id,
            ApiField.FILTER: filters or [],
            ApiField.TYPE: collection_type,
        }
        if with_meta:
            data.update({ApiField.EXTRA_FIELDS: [ApiField.META]})
        return self.get_list_all_pages(method, data)

    def get_info_by_id(self, id: int, with_meta: bool = False) -> EntitiesCollectionInfo:
        """
        Get information about Entities Collection with given ID.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param with_meta: If True, includes meta information in the response. Defaults to False.
        :type with_meta: bool, optional
        :return: Information about Entities Collection.
        :rtype: :class:`EntitiesCollectionInfo`
        :Usage example:

            .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection = api.entities_collection.get_info_by_id(2)
            print(collection)
            # Output:
            # {
            #     "id": 1,
            #     "teamId": 1,
            #     "projectId": 1,
            #     "name": "ds",
            #     "description": "",
            #     "createdAt": "2018-08-21T14:25:56.140Z",
            #     "updatedAt": "2018-08-21T14:25:56.140Z"
            # }
        """
        method = "entities-collections.info"
        if with_meta:
            extra_fields = [ApiField.META]
        else:
            extra_fields = None
        return self._get_info_by_id(id, method, extra_fields)

    def get_info_by_ai_search_key(
        self, project_id: int, ai_search_key: str
    ) -> Optional[EntitiesCollectionInfo]:
        """
        Get information about Entities Collection of type `CollectionType.AI_SEARCH` with given AI search key.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param ai_search_key: AI search key for the collection.
        :type  ai_search_key: str
        :return: Information about Entities Collection.
        :rtype: :class:`EntitiesCollectionInfo`
        :Usage example:
         .. code-block:: python
            import supervisely as sly
            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()
            project_id = 123
            ai_search_key = "0ed6a5256433bbe32822949d563d476a"

            # Get collection by AI search key
            collection = api.entities_collection.get_info_by_ai_search_key(project_id, ai_search_key)
        """
        collections = self.get_list(
            project_id=project_id,
            with_meta=True,
            collection_type=CollectionType.AI_SEARCH,
        )
        return next(
            (collection for collection in collections if collection.ai_search_key == ai_search_key),
            None,
        )

    def _get_response_by_id(self, id, method, id_field, extra_fields=None):
        """Differs from the original method by using extra_fields."""
        try:
            data = {id_field: id}
            if extra_fields is not None:
                data.update({ApiField.EXTRA_FIELDS: extra_fields})
            return self._api.post(method, data)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            else:
                raise error

    def _get_info_by_id(self, id, method, extra_fields=None):
        """_get_info_by_id"""
        response = self._get_response_by_id(
            id, method, id_field=ApiField.ID, extra_fields=extra_fields
        )
        return self._convert_json_info(response.json()) if (response is not None) else None

    def _get_update_method(self):
        """ """
        return "entities-collections.editInfo"

    def _remove_api_method_name(self):
        """ """
        return "entities-collections.remove"

    def add_items(self, id: int, items: List[Union[int, CollectionItem]]) -> List[Dict[str, int]]:
        """
        Add items to Entities Collection.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param items: List of items to add to the collection. Could be a list of entity IDs (int) or CollectionItem objects.
        :type items: List[Union[int, CollectionItem]]
        :return: List of added items with their IDs and creation timestamps.
        :rtype: List[Dict[str, int]]
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.api.entities_collection_api import CollectionItem

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection_id = 2
            item_ids = [525, 526]
            items = [CollectionItem(entity_id=item_id) for item_id in item_ids]
            new_items = api.entities_collection.add_items(collection_id, items)
            print(new_items)
            # Output: [
            #   {"id": 1, "entityId": 525, 'createdAt': '2025-04-10T08:49:41.852Z'},
            #   {"id": 2, "entityId": 526, 'createdAt': '2025-04-10T08:49:41.852Z'}
            ]
        """
        if all(isinstance(item, int) for item in items):
            data = {ApiField.COLLECTION_ID: id, ApiField.ENTITY_IDS: items}
        elif all(isinstance(item, CollectionItem) for item in items):
            items = [item.to_json() for item in items]
            data = {ApiField.COLLECTION_ID: id, ApiField.ENTITY_ITEMS: items}
        else:
            raise ValueError(
                "Items list must contain only integers or only CollectionItem instances, not a mix."
            )

        response = self._api.post("entities-collections.items.bulk.add", data)
        response = response.json()
        if len(response["missing"]) > 0:
            raise RuntimeError(
                f"Failed to add items to Entities Collection. IDs: {response['missing']}. "
            )
        return response["items"]

    def get_items(
        self,
        collection_id: int,
        collection_type: CollectionTypeFilter,
        project_id: Optional[int] = None,
        ai_search_threshold: Optional[float] = None,
        ai_search_threshold_direction: AiSearchThresholdDirection = AiSearchThresholdDirection.ABOVE,
    ) -> List[ImageInfo]:
        """
        Get items from Entities Collection.

        :param collection_id: Entities Collection ID in Supervisely.
        :type collection_id: int
        :param collection_type: Type of the collection. Can be CollectionTypeFilter.AI_SEARCH or CollectionTypeFilter.DEFAULT.
        :type collection_type: CollectionTypeFilter
        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :param ai_search_threshold: AI search threshold for filtering items. Optional, defaults to None.
        :type ai_search_threshold: float, optional
        :param ai_search_threshold_direction: Direction for the AI search threshold. Optional, defaults to 'above'.
        :type ai_search_threshold_direction: str
        :return: List of ImageInfo objects.
        :rtype: List[ImageInfo]
        :raises RuntimeError: If Entities Collection with given ID not found.
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection_id = 2
            project_id = 4
            item_ids = api.entities_collection.get_items(collection_id, project_id)
            print(item_ids)
        """
        if project_id is None:
            info = self.get_info_by_id(collection_id)
            if info is None:
                raise RuntimeError(f"Entities Collection with id={collection_id} not found.")
            project_id = info.project_id

        if collection_type == CollectionTypeFilter.AI_SEARCH:
            return self._api.image.get_list(
                project_id=project_id,
                ai_search_collection_id=collection_id,
                extra_fields=[ApiField.EMBEDDINGS_UPDATED_AT],
                ai_search_threshold=ai_search_threshold,
                ai_search_threshold_direction=ai_search_threshold_direction,
            )
        else:
            return self._api.image.get_list(
                project_id=project_id,
                entities_collection_id=collection_id,
            )

    def remove_items(self, id: int, items: List[int]) -> List[Dict[str, int]]:
        """
        Remove items from Entities Collection.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param items: List of entity IDs in Supervisely, e.g. image IDs etc.
        :type items: List[int]
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            api = sly.Api.from_env()

            collection_id = 2
            item_ids = [525, 526, 527]
            removed_items = api.entities_collection.remove_items(collection_id, item_ids)
            # print(removed_items)
            # Output: [{"id": 1, "entityId": 525}, {"id": 2, "entityId": 526}]
        """
        method = "entities-collections.items.bulk.remove"

        data = {ApiField.COLLECTION_ID: id, ApiField.ENTITY_IDS: items}
        response = self._api.post(method, data)
        response = response.json()
        if len(response["missing"]) > 0:
            logger.warning(
                f"Failed to remove items from Entities Collection. IDs: {response['missing']}. "
            )
        return response["items"]



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from supervisely.project.versioning.schema_fields import VersionSchemaField


@dataclass(frozen=True)
class VideoSnapshotSchema:
    schema_version: str

    def datasets_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_DATASET_ID, pa_module.int64()),
                (VersionSchemaField.PARENT_SRC_DATASET_ID, pa_module.int64()),
                (VersionSchemaField.NAME, pa_module.utf8()),
                (VersionSchemaField.FULL_PATH, pa_module.utf8()),
                (VersionSchemaField.DESCRIPTION, pa_module.utf8()),
                (VersionSchemaField.CUSTOM_DATA, pa_module.utf8()),
            ]
        )

    def videos_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_VIDEO_ID, pa_module.int64()),
                (VersionSchemaField.SRC_DATASET_ID, pa_module.int64()),
                (VersionSchemaField.NAME, pa_module.utf8()),
                (VersionSchemaField.HASH, pa_module.utf8()),
                (VersionSchemaField.LINK, pa_module.utf8()),
                (VersionSchemaField.FRAMES_COUNT, pa_module.int32()),
                (VersionSchemaField.FRAME_WIDTH, pa_module.int32()),
                (VersionSchemaField.FRAME_HEIGHT, pa_module.int32()),
                (VersionSchemaField.FRAMES_TO_TIMECODES, pa_module.utf8()),
                (VersionSchemaField.META, pa_module.utf8()),
                (VersionSchemaField.CUSTOM_DATA, pa_module.utf8()),
                (VersionSchemaField.CREATED_AT, pa_module.utf8()),
                (VersionSchemaField.UPDATED_AT, pa_module.utf8()),
                (VersionSchemaField.ANN_JSON, pa_module.utf8()),
            ]
        )

    def objects_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_OBJECT_ID, pa_module.int64()),
                (VersionSchemaField.SRC_VIDEO_ID, pa_module.int64()),
                (VersionSchemaField.CLASS_NAME, pa_module.utf8()),
                (VersionSchemaField.KEY, pa_module.utf8()),
                (VersionSchemaField.TAGS_JSON, pa_module.utf8()),
            ]
        )

    def figures_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_FIGURE_ID, pa_module.int64()),
                (VersionSchemaField.SRC_OBJECT_ID, pa_module.int64()),
                (VersionSchemaField.SRC_VIDEO_ID, pa_module.int64()),
                (VersionSchemaField.FRAME_INDEX, pa_module.int32()),
                (VersionSchemaField.GEOMETRY_TYPE, pa_module.utf8()),
                (VersionSchemaField.GEOMETRY_JSON, pa_module.utf8()),
            ]
        )

    def dataset_row(
        self,
        *,
        src_dataset_id: int,
        parent_src_dataset_id: Optional[int],
        name: str,
        full_path: str,
        description: Optional[str],
        custom_data: Optional[dict],
    ) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_DATASET_ID: src_dataset_id,
            VersionSchemaField.PARENT_SRC_DATASET_ID: parent_src_dataset_id,
            VersionSchemaField.NAME: name,
            VersionSchemaField.FULL_PATH: full_path,
            VersionSchemaField.DESCRIPTION: description,
            VersionSchemaField.CUSTOM_DATA: (
                json.dumps(custom_data) if isinstance(custom_data, dict) and len(custom_data) > 0 else None
            ),
        }

    def video_row(
        self,
        *,
        src_video_id: int,
        src_dataset_id: int,
        name: str,
        hash: Optional[str],
        link: Optional[str],
        frames_count: Optional[int],
        frame_width: Optional[int],
        frame_height: Optional[int],
        frames_to_timecodes: Optional[list],
        meta: Optional[dict],
        custom_data: Optional[dict],
        created_at: Optional[str],
        updated_at: Optional[str],
        ann_json: dict,
    ) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_VIDEO_ID: src_video_id,
            VersionSchemaField.SRC_DATASET_ID: src_dataset_id,
            VersionSchemaField.NAME: name,
            VersionSchemaField.HASH: hash,
            VersionSchemaField.LINK: link,
            VersionSchemaField.FRAMES_COUNT: frames_count,
            VersionSchemaField.FRAME_WIDTH: frame_width,
            VersionSchemaField.FRAME_HEIGHT: frame_height,
            VersionSchemaField.FRAMES_TO_TIMECODES: (
                json.dumps(frames_to_timecodes) if frames_to_timecodes else None
            ),
            VersionSchemaField.META: json.dumps(meta) if meta else None,
            VersionSchemaField.CUSTOM_DATA: json.dumps(custom_data) if custom_data else None,
            VersionSchemaField.CREATED_AT: created_at,
            VersionSchemaField.UPDATED_AT: updated_at,
            VersionSchemaField.ANN_JSON: json.dumps(ann_json),
        }

    def object_row(
        self,
        *,
        src_object_id: int,
        src_video_id: int,
        class_name: str,
        key_hex: str,
        tags_json: Optional[List[Dict[str, Any]]],
    ) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_OBJECT_ID: src_object_id,
            VersionSchemaField.SRC_VIDEO_ID: src_video_id,
            VersionSchemaField.CLASS_NAME: class_name,
            VersionSchemaField.KEY: key_hex,
            VersionSchemaField.TAGS_JSON: json.dumps(tags_json) if tags_json is not None else None,
        }

    def figure_row(
        self,
        *,
        src_figure_id: int,
        src_object_id: int,
        src_video_id: int,
        frame_index: int,
        geometry_type: str,
        geometry_json: dict,
    ) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_FIGURE_ID: src_figure_id,
            VersionSchemaField.SRC_OBJECT_ID: src_object_id,
            VersionSchemaField.SRC_VIDEO_ID: src_video_id,
            VersionSchemaField.FRAME_INDEX: frame_index,
            VersionSchemaField.GEOMETRY_TYPE: geometry_type,
            VersionSchemaField.GEOMETRY_JSON: json.dumps(geometry_json),
        }

    def dataset_row_from_ds_info(self, ds_info, *, full_path: str, custom_data: Optional[dict]):
        return self.dataset_row(
            src_dataset_id=ds_info.id,
            parent_src_dataset_id=ds_info.parent_id,
            name=ds_info.name,
            full_path=full_path,
            description=getattr(ds_info, "description", None),
            custom_data=custom_data,
        )

    def video_row_from_video_info(self, video_info, *, src_dataset_id: int, ann_json: dict):
        return self.video_row(
            src_video_id=video_info.id,
            src_dataset_id=src_dataset_id,
            name=video_info.name,
            hash=getattr(video_info, "hash", None),
            link=getattr(video_info, "link", None),
            frames_count=getattr(video_info, "frames_count", None),
            frame_width=getattr(video_info, "frame_width", None),
            frame_height=getattr(video_info, "frame_height", None),
            frames_to_timecodes=getattr(video_info, "frames_to_timecodes", None),
            meta=getattr(video_info, "meta", None),
            custom_data=getattr(video_info, "custom_data", None),
            created_at=getattr(video_info, "created_at", None),
            updated_at=getattr(video_info, "updated_at", None),
            ann_json=ann_json,
        )

    def object_row_from_object(self, obj, *, src_object_id: int, src_video_id: int) -> Dict[str, Any]:
        return self.object_row(
            src_object_id=src_object_id,
            src_video_id=src_video_id,
            class_name=obj.obj_class.name,
            key_hex=obj.key().hex,
            tags_json=obj.tags.to_json() if getattr(obj, "tags", None) is not None else None,
        )

    def figure_row_from_figure(
        self,
        fig,
        *,
        figure_row_idx: int,
        src_object_id: int,
        src_video_id: int,
        frame_index: int,
    ) -> Dict[str, Any]:
        return self.figure_row(
            src_figure_id=figure_row_idx + 1,
            src_object_id=src_object_id,
            src_video_id=src_video_id,
            frame_index=frame_index,
            geometry_type=fig.geometry.geometry_name(),
            geometry_json=fig.geometry.to_json(),
        )

_VIDEO_SCHEMAS: Dict[str, VideoSnapshotSchema] = {
    "v2.0.0": VideoSnapshotSchema(schema_version="v2.0.0"),
}




import imghdr
import os
from typing import List

import supervisely.convert.pointcloud.ply.ply_helper as ply_helper
from supervisely import PointcloudAnnotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.io.fs import JUNK_FILES, get_file_ext, get_file_name


class PlyConverter(PointcloudConverter):

    def __str__(self) -> str:
        return AvailablePointcloudConverters.PLY

    @property
    def ann_ext(self) -> str:
        return ".json"

    def validate_format(self) -> bool:
        ply_list = []
        ann_dict = {}
        rimg_dict = {}
        used_img_ext = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if file in JUNK_FILES:
                    continue
                elif ext == ".ply":
                    ply_list.append(full_path)
                elif ext ==  self.ann_ext:
                    ann_dict[file] = full_path
                elif imghdr.what(full_path):
                    rimg_dict[file] = full_path
                    if ext not in used_img_ext:
                        used_img_ext.append(ext)

        # create Items
        self._items = []
        for ply_path in ply_list:
            pcd_path = ply_path.replace(".ply", ".pcd")
            ply_helper.ply2pcd(ply_path, pcd_path)
            if not os.path.exists(pcd_path):
                logger.warn(f"Failed to convert PLY to PCD. Skipping: {ply_path}")
                continue
            item = self.Item(pcd_path)
            for ext in used_img_ext:
                rimg_name = f"{item.name}{ext}"
                if not rimg_name in rimg_dict:
                    rimg_name = f"{get_file_name(item.name)}{ext}"
                if rimg_name in rimg_dict:
                    rimg_path = rimg_dict[rimg_name]
                    rimg_ann_name = f"{rimg_name}.json"
                    if rimg_ann_name in ann_dict:
                        rimg_ann_path = ann_dict[rimg_ann_name]
                        item.set_related_images((rimg_path, rimg_ann_path))
            self._items.append(item)
        return self.items_count > 0

    def to_supervisely(
        self,
        item: PointcloudConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> PointcloudAnnotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()



# coding: utf-8
from __future__ import annotations

import asyncio
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.io.json import load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.video_tag_collection import VideoTagCollection


class VideoAnnotationAPI(EntityAnnotationAPI):
    """
    :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` for a single video. :class:`VideoAnnotationAPI<VideoAnnotationAPI>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        video_id = 186648102
        ann_info = api.video.annotation.download(video_id)
    """

    _method_download_bulk = "videos.annotations.bulk.info"
    _entity_ids_str = ApiField.VIDEO_IDS

    def download(self, video_id: int) -> Dict:
        """
        Download information about VideoAnnotation by video ID from API.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :return: Information about VideoAnnotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 198702499
            ann_info = api.video.annotation.download(video_id)
            print(ann_info)
            # Output: {
            #     "videoId": 198702499,
            #     "videoName": "Videos_dataset_cars_cars.mp4",
            #     "createdAt": "2021-03-23T13:14:25.536Z",
            #     "updatedAt": "2021-03-23T13:16:43.300Z",
            #     "description": "",
            #     "tags": [],
            #     "objects": [],
            #     "size": {
            #         "height": 2160,
            #         "width": 3840
            #     },
            #     "framesCount": 326,
            #     "frames": []
            # }
        """

        video_info = self._api.video.get_info_by_id(video_id)
        return self._download(video_info.dataset_id, video_id)

    def append(
        self,
        video_id: int,
        ann: VideoAnnotation,
        key_id_map: Optional[KeyIdMap] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Loads an VideoAnnotation to a given video ID in the API.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param ann: VideoAnnotation object.
        :type ann: VideoAnnotation
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :param progress: Progress.
        :type progress: Optional[Union[tqdm, Callable]]
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 198704259
            api.video.annotation.append(video_id, video_ann)
        """

        info = self._api.video.get_info_by_id(video_id)
        self._append(
            self._api.video.tag,
            self._api.video.object,
            self._api.video.figure,
            info.project_id,
            info.dataset_id,
            video_id,
            ann.tags,
            ann.objects,
            ann.figures,
            key_id_map,
            progress_cb,
        )

    def upload_paths(
        self,
        video_ids: List[int],
        ann_paths: List[str],
        project_meta: ProjectMeta,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Loads an VideoAnnotations from a given paths to a given videos IDs in the API. Videos IDs must be from one dataset.

        :param video_ids: Videos IDs in Supervisely.
        :type video_ids: List[int]
        :param ann_paths: Paths to annotations on local machine.
        :type ann_paths: List[str]
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>` for VideoAnnotations.
        :type project_meta: ProjectMeta
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_ids = [121236918, 121236919]
            ann_paths = ['/home/admin/work/supervisely/example/ann1.json', '/home/admin/work/supervisely/example/ann2.json']
            api.video.annotation.upload_paths(video_ids, ann_paths, meta)
        """
        # video_ids from the same dataset
        for video_id, ann_path in zip(video_ids, ann_paths):
            ann_json = load_json_file(ann_path)
            ann = VideoAnnotation.from_json(ann_json, project_meta)

            # ignore existing key_id_map because the new objects will be created
            self.append(video_id, ann)
            if progress_cb is not None:
                progress_cb(1)

    def upload_paths_multiview(
        self,
        video_ids: List[int],
        ann_paths: List[str],
        project_meta: ProjectMeta,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Upload VideoAnnotations for multi-view video project.
        All provided video ids must belong to the same project and dataset.

        Objects with the same key are created only once and shared between videos.
        In this mode annotation objects are created without binding to a specific entityId.

        :param video_ids: Video IDs in Supervisely.
        :type video_ids: List[int]
        :param ann_paths: Paths to annotations on local machine.
        :type ann_paths: List[str]
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>` for VideoAnnotations.
        :type project_meta: ProjectMeta
        :param progress_cb: Function for tracking upload progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`
        """
        if len(video_ids) != len(ann_paths):
            raise RuntimeError(
                f'Can not match "video_ids" and "ann_paths" lists, len(video_ids) != len(ann_paths): {len(video_ids)} != {len(ann_paths)}'
            )
        if len(video_ids) == 0:
            return

        anns = []
        for ann_path in ann_paths:
            ann_json = load_json_file(ann_path)
            ann = VideoAnnotation.from_json(ann_json, project_meta)
            anns.append(ann)

        self.upload_anns_multiview(video_ids, anns, progress_cb)

    def upload_anns_multiview(
        self,
        video_ids: List[int],
        anns: List[VideoAnnotation],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> None:
        """
        Upload already constructed VideoAnnotation objects for multi-view video project.
        All provided video ids must belong to the same project and dataset.

        Objects with the same key are created only once and shared between videos.
        In this mode annotation objects are created without binding to a specific entityId.

        :param video_ids: Video IDs in Supervisely.
        :type video_ids: List[int]
        :param anns: List of VideoAnnotation objects corresponding to the video_ids.
        :type anns: List[VideoAnnotation]
        :param progress_cb: Function for tracking upload progress (by number of figures).
        :type progress_cb: tqdm or callable, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: None
        :rtype: :class:`NoneType`
        """
        if len(video_ids) != len(anns):
            raise RuntimeError(
                'Can not match "video_ids" and "anns" lists, len(video_ids) != len(anns)'
            )
        if len(video_ids) == 0:
            return

        try:
            video_infos = self._api.video.get_info_by_id_batch(video_ids)
        except RuntimeError as e:
            raise RuntimeError("All videos must belong to the same project and dataset.") from e

        project_id = video_infos[0].project_id
        dataset_id = video_infos[0].dataset_id

        tag_api = self._api.video.tag
        object_api = self._api.video.object
        figure_api = self._api.video.figure

        key_id_map = key_id_map or KeyIdMap()
        for video_id, ann in zip(video_ids, anns):
            tag_api.append_to_entity(video_id, project_id, ann.tags, key_id_map=key_id_map)
            new_objects = []
            existing_objects = []
            for obj in ann.objects:
                if key_id_map.get_object_id(obj.key()) is None:
                    new_objects.append(obj)
                else:
                    existing_objects.append(obj)
            if len(new_objects) > 0:
                object_api._append_bulk(
                    tag_api=tag_api,
                    entity_id=video_id,
                    project_id=project_id,
                    dataset_id=dataset_id,
                    objects=new_objects,
                    key_id_map=key_id_map,
                    is_pointcloud=False,
                    is_video_multi_view=True,
                )
            tags_to_obj = {}
            for obj in existing_objects:
                obj_id = key_id_map.get_object_id(obj.key())
                tags_to_obj[obj_id] = obj.tags
            if len(tags_to_obj) > 0:
                tag_api.add_tags_collection_to_objects(
                    project_id, tags_to_obj, is_video_multi_view=True, entity_id=video_id
                )

            figure_api.append_bulk(video_id, ann.figures, key_id_map)
            if progress_cb is not None and len(ann.figures) > 0:
                if hasattr(progress_cb, "update") and callable(getattr(progress_cb, "update")):
                    progress_cb.update(len(ann.figures))
                else:
                    progress_cb(len(ann.figures))

    def copy_batch(
        self,
        src_video_ids: List[int],
        dst_video_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Copy annotations from one images IDs to another in API.

        :param src_video_ids: Images IDs in Supervisely.
        :type src_video_ids: List[int]
        :param dst_video_ids: Unique IDs of images in API.
        :type dst_video_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :raises: :class:`RuntimeError`, if len(src_video_ids) != len(dst_video_ids)
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from tqdm import tqdm

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_ids = [121236918, 121236919]
            dst_ids = [547837053, 547837054]
            p = tqdm(desc="Annotations copy: ", total=len(src_ids))

            copy_anns = api.annotation.copy_batch(src_ids, dst_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:24:31.286Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:24:31.288Z", "level": "info"}
        """
        if len(src_video_ids) != len(dst_video_ids):
            raise RuntimeError(
                'Can not match "src_video_ids" and "dst_video_ids" lists, '
                "len(src_video_ids) != len(dst_video_ids)"
            )
        if len(src_video_ids) == 0:
            return

        src_dataset_id = self._api.video.get_info_by_id(src_video_ids[0]).dataset_id
        dst_dataset_id = self._api.video.get_info_by_id(dst_video_ids[0]).dataset_id
        dst_dataset_info = self._api.dataset.get_info_by_id(dst_dataset_id)
        dst_project_meta = ProjectMeta.from_json(
            self._api.project.get_meta(dst_dataset_info.project_id)
        )
        for src_ids_batch, dst_ids_batch in zip(batched(src_video_ids), batched(dst_video_ids)):
            ann_jsons = self.download_bulk(src_dataset_id, src_ids_batch)
            for dst_id, ann_json in zip(dst_ids_batch, ann_jsons):
                try:
                    ann = VideoAnnotation.from_json(
                        ann_json, dst_project_meta, key_id_map=KeyIdMap()
                    )
                except Exception as e:
                    raise RuntimeError("Failed to validate Annotation") from e
                self.append(dst_id, ann)
                if progress_cb is not None:
                    progress_cb(1)

    async def download_async(
        self,
        video_id: int,
        video_info=None,
        semaphore: Optional[asyncio.Semaphore] = None,
        force_metadata_for_links: bool = True,
        integer_coords: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Dict:
        """
        Download information about VideoAnnotation by video ID from API asynchronously.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param video_info: Does not affect the result, but is left for compatibility with the method signature.
        :type video_info: VideoInfo, optional
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param force_metadata_for_links: If True, updates meta for videos with links.
        :type force_metadata_for_links: bool, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param progress_cb: Progress callback to track download progress.
        :type progress_cb: Union[tqdm, Callable], optional
        :return: Information about VideoAnnotation in json format
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 198702499
            loop = sly.utils.get_or_create_event_loop()
            ann_info = loop.run_until_complete(api.video.annotation.download_async(video_id))
        """
        return await self.download_bulk_async(
            video_ids=[video_id],
            semaphore=semaphore,
            force_metadata_for_links=force_metadata_for_links,
            integer_coords=integer_coords,
            progress_cb=progress_cb,
        )

    async def download_bulk_async(
        self,
        video_ids: List[int],
        semaphore: Optional[asyncio.Semaphore] = None,
        force_metadata_for_links: bool = True,
        integer_coords: bool = True,
        batch_size: int = 10,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[Dict]:
        """
        Download information about VideoAnnotation in bulk by video IDs from API asynchronously.

        :param video_ids: List of Video IDs in Supervisely. All videos must be from the same dataset.
        :type video_ids: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param force_metadata_for_links: If True, updates meta for videos with links.
        :type force_metadata_for_links: bool, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param batch_size: Batch size for parallel downloads. Default is 10 as an optimal value.
        :type batch_size: int, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about VideoAnnotations in json format
        :rtype: :class:`list`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_ids = [198702499, 198702500, 198702501]
            loop = sly.utils.get_or_create_event_loop()
            ann_infos = loop.run_until_complete(api.video.annotation.download_bulk_async(video_ids))
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def fetch_with_semaphore(batch):
            async with semaphore:
                json_data = {
                    self._entity_ids_str: batch,
                    ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
                    ApiField.INTEGER_COORDS: integer_coords,
                }
                response = await self._api.post_async(
                    self._method_download_bulk,
                    json=json_data,
                )
                if progress_cb is not None:
                    progress_cb(len(batch))
                return response.json()

        tasks = [fetch_with_semaphore(batch) for batch in batched(video_ids, batch_size=batch_size)]
        responses = await asyncio.gather(*tasks)
        json_response = [item for response in responses for item in response]
        return json_response



# coding: utf-8

import uuid
from typing import Optional, Union

from numpy import ndarray

from supervisely.geometry.mask_3d import Mask3D
from supervisely.video_annotation.video_object import VideoObject
from supervisely.volume_annotation import volume_figure
from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection


class VolumeObject(VideoObject):
    """
    VolumeObject object for :class:`VolumeAnnotation<supervisely.volume_annotation.volume_annotation.VolumeAnnotation>`. :class:`VolumeObject<VolumeObject>` object is immutable.

    :param obj_class: VolumeObject :class:`class<supervisely.annotation.obj_class.ObjClass>`.
    :type obj_class: ObjClass
    :param tags: VolumeObject :class:`tags<supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection>`.
    :type tags: VolumeTagCollection, optional
    :param key: The UUID key associated with the VolumeFigure.
    :type key: UUID, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which VolumeObject belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created VolumeObject.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when VolumeObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when VolumeObject was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :param mask_3d: Path for local geometry file, array with geometry data or Mask3D geometry object
    :type mask_3d: Union[str, ndarray, Mask3D], optional
    :param custom_data: Custom data associated with the VolumeObject.
    :type custom_data: dict, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
        volume_obj_heart = sly.VolumeObject(obj_class_heart)
        volume_obj_heart_json = volume_obj_heart.to_json()
        print(volume_obj_heart_json)
        # Output: {
        #     "key": "6b819f1840f84d669b32cdec225385f0",
        #     "classTitle": "heart",
        #     "tags": []
        # }
    """

    def __init__(
        self,
        obj_class,
        tags: Optional[VolumeTagCollection] = None,
        key: Optional[uuid.UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        mask_3d: Optional[Union[Mask3D, ndarray, str]] = None,
        custom_data: Optional[dict] = None,
    ):
        super().__init__(
            obj_class=obj_class,
            tags=tags,
            key=key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        if mask_3d is not None:
            if isinstance(mask_3d, str):
                self.figure = volume_figure.VolumeFigure(
                    self,
                    geometry=Mask3D.create_from_file(mask_3d),
                    labeler_login=labeler_login,
                    updated_at=updated_at,
                    created_at=created_at,
                    custom_data=custom_data,
                )
            elif isinstance(mask_3d, ndarray):
                self.figure = volume_figure.VolumeFigure(
                    self,
                    geometry=Mask3D(mask_3d),
                    labeler_login=labeler_login,
                    updated_at=updated_at,
                    created_at=created_at,
                    custom_data=custom_data,
                )
            elif isinstance(mask_3d, Mask3D):
                self.figure = volume_figure.VolumeFigure(
                    self,
                    geometry=mask_3d,
                    labeler_login=labeler_login,
                    updated_at=updated_at,
                    created_at=created_at,
                    custom_data=custom_data,
                )
            else:
                raise TypeError("mask_3d type must be one of [Mask3D, ndarray, str]")
